# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC, abstractmethod
from uuid import UUID

import numpy as np

from ..data import Data, DataAssociationEnum
from ..groups import PropertyGroup
from ..shared.utils import box_intersect, mask_by_extent
from .points import Points


class CellObject(Points, ABC):
    """
    Base class for object with cells.

    :param cells: Array of indices defining connecting vertices.
    """

    _attribute_map: dict = Points._attribute_map.copy()
    _TYPE_UID: UUID | None = None

    def __init__(
        self,
        cells: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self._cells = self.validate_cells(cells)

    @property
    def cells(self) -> np.ndarray:
        """
        Array of indices defining connecting vertices.
        """
        if self._cells is None and self.on_file:
            self._cells = self.workspace.fetch_array_attribute(self, "cells")

        return self._cells

    @cells.setter
    def cells(self, cells: np.ndarray | list | tuple):
        cells = self.validate_cells(cells)
        if self._cells is not None and self._cells.shape != cells.shape:
            raise ValueError(
                "New cells array must have the same shape as the current cells array."
            )
        self._cells = cells

        self.workspace.update_attribute(self, "cells")

    @property
    def centroids(self) -> np.ndarray | None:
        """
        Compute the centroids of the cells.
        """
        return np.mean(self.vertices[self.cells], axis=1)

    @property
    def locations(self):
        return self.vertices

    def mask_by_extent(
        self,
        extent: np.ndarray,
        inverse: bool = False,
    ) -> np.ndarray | None:
        """
        Extension of :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.
        """
        if self.extent is None or not box_intersect(self.extent, extent):
            return None

        vert_mask = mask_by_extent(self.vertices, extent, inverse=inverse)

        # Check for orphan vertices
        cell_mask = np.all(vert_mask[self.cells], axis=1)
        orphan_mask = np.zeros_like(vert_mask, dtype=bool)
        orphan_mask[self.cells[cell_mask].flatten()] = True
        vert_mask &= orphan_mask

        if ~np.any(vert_mask):
            return None

        return vert_mask

    @property
    def n_cells(self) -> int:
        """
        Number of vertices
        """
        return self.cells.shape[0]

    def remove_cells(self, indices: list[int] | np.ndarray, clear_cache: bool = False):
        """
        Safely remove cells and corresponding data entries.

        :param indices: Indices of cells to be removed.
        :param clear_cache: Clear cache of data values.
        """
        if isinstance(indices, (list, tuple)):
            indices = np.array(indices)

        if not isinstance(indices, np.ndarray):
            raise TypeError("Indices must be a list or numpy array.")

        if (
            isinstance(self.cells, np.ndarray)
            and np.max(indices) > self.cells.shape[0] - 1
        ):
            raise ValueError("Found indices larger than the number of cells.")

        cells = np.delete(self.cells, indices, axis=0)
        self.load_children_values()
        self._cells = self.validate_cells(cells)
        self._remove_children_values(
            indices, DataAssociationEnum.CELL, clear_cache=clear_cache
        )
        self.workspace.update_attribute(self, "cells")

    def remove_vertices(
        self, indices: list[int] | np.ndarray, clear_cache: bool = False
    ):
        """
        Safely remove vertices and cells and corresponding data entries.

        :param indices: Indices of vertices to be removed.
        :param clear_cache: Clear cache of data values.
        """
        n_vertices = self.vertices.shape[0]

        super().remove_vertices(indices, clear_cache=clear_cache)

        vert_index = np.ones(n_vertices, dtype=bool)
        vert_index[indices] = False
        new_index = np.ones_like(vert_index, dtype=int)
        new_index[vert_index] = np.arange(self.vertices.shape[0])

        cell_mask = np.where(~np.all(vert_index[self.cells], axis=1))
        self._cells = new_index[self.cells]
        self.remove_cells(cell_mask)

        self.workspace.update_attribute(self, "cells")

    def copy(  # pylint: disable=too-many-branches
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        cell_mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Sub-class extension of :func:`~geoh5py.objects.points.Points.copy`.

        Additions:
            cell_mask: Array of indices to sub-sample the input entity cells.
        """
        if mask is not None and self.vertices is not None:
            if not isinstance(mask, np.ndarray) or mask.shape != (
                self.vertices.shape[0],
            ):
                raise ValueError("Mask must be an array of shape (n_vertices,).")

            kwargs.update({"vertices": self.vertices[mask, :]})

        new_cells = getattr(self, "cells", None)
        if mask is not None:
            new_id = np.ones_like(mask, dtype=int)
            new_id[mask] = np.arange(np.sum(mask))

            if cell_mask is None:
                cell_mask = np.all(mask[self.cells], axis=1)

            new_cells = new_id[self.cells]

        if cell_mask is not None and new_cells is not None:
            new_cells = new_cells[cell_mask, :]
            kwargs.update(
                {
                    "cells": new_cells,
                }
            )

        new_object = super().copy(
            parent=parent,
            copy_children=False,
            clear_cache=clear_cache,
            mask=mask,
            **kwargs,
        )

        if copy_children:
            children_map = {}
            for child in self.children:
                if isinstance(child, PropertyGroup):
                    continue
                if isinstance(child, Data):
                    if child.name in ["A-B Cell ID"]:
                        continue

                    child_mask = mask
                    if (
                        child.association is DataAssociationEnum.CELL
                        and cell_mask is not None
                    ):
                        child_mask = cell_mask
                    elif child.association is not DataAssociationEnum.VERTEX:
                        child_mask = None

                    child_copy = child.copy(
                        parent=new_object,
                        clear_cache=clear_cache,
                        mask=child_mask,
                    )
                else:
                    child_copy = self.workspace.copy_to_parent(
                        child, new_object, clear_cache=clear_cache
                    )
                children_map[child.uid] = child_copy.uid

            if self.property_groups:
                self.workspace.copy_property_groups(
                    new_object, self.property_groups, children_map
                )

        return new_object

    @abstractmethod
    def validate_cells(self, indices: list | tuple | np.ndarray | None) -> np.ndarray:
        """
        Validate or generate cells defining the connection between vertices.

        :param indices: Array of indices. If None provided, the
            vertices are connected sequentially.

        :return: Array of indices defining connected vertices.
        """



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

import uuid

from .points import Points
from .surface import Surface


class IntegratorPoints(Points):
    """
    INTEGRATOR Points object.
    Sub-class of :obj:`geoh5py.objects.points.Points`.
    """

    _TYPE_UID = uuid.UUID("{6832ACF3-78AA-44D3-8506-9574A3510C44}")
    _default_name = "Geoscience INTEGRATOR Points"


class NeighbourhoodSurface(Surface):
    """
    Points object made up of vertices.
    """

    _TYPE_UID = uuid.UUID("{88087FB8-76AE-445B-9CDF-68DBCE530404}")
    _default_name = "Neighbourhood Surface"



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

import numpy as np

from geoh5py.data import Data
from geoh5py.groups.property_group import GroupTypeEnum, PropertyGroup


if TYPE_CHECKING:
    from .concatenator import Concatenator
    from .object import ConcatenatedObject


class ConcatenatedPropertyGroup(PropertyGroup):
    def __init__(self, parent: ConcatenatedObject, **kwargs):
        super().__init__(parent, **kwargs)

    @property
    def concatenator(self) -> Concatenator:
        """
        Parental Concatenator entity.
        """
        return self._parent.concatenator

    def is_collocated(
        self,
        locations: np.ndarray,
        collocation_distance: float,
    ) -> bool:
        """
        True if locations are collocated with property group.

        :param locations: Locations to check.
        :param collocation_distance: tolerance for similarity check.
        """

        if (
            self.locations is None
            or locations.ndim != self.locations.ndim
            or len(locations) != len(self.locations)
            or not np.allclose(locations, self.locations, atol=collocation_distance)
        ):
            return False

        return True

    @property
    def locations(self) -> np.ndarray:
        """Return depths or intervals array if either exists else None."""

        if self.depth_ is not None:
            return self.depth_.values

        if self.from_ is not None and self.to_ is not None:
            return np.c_[self.from_.values, self.to_.values]

        return None

    @property
    def depth_(self):
        if (
            self.properties is None
            or len(self.properties) < 1
            or self._property_group_type != GroupTypeEnum.DEPTH
        ):
            return None

        data = self.parent.get_data(  # pylint: disable=no-value-for-parameter
            self.properties[0]
        )

        if any(data) and isinstance(data[0], Data) and "depth" in data[0].name.lower():
            return data[0]

        return None

    @property
    def from_(self):
        """Return the data entities defined the 'from' depth intervals."""
        if (
            self.properties is None
            or len(self.properties) < 1
            or self._property_group_type != GroupTypeEnum.INTERVAL
        ):
            return None

        data = self.parent.get_data(  # pylint: disable=no-value-for-parameter
            self.properties[0]
        )

        if any(data) and isinstance(data[0], Data) and "from" in data[0].name.lower():
            return data[0]

        return None

    @property
    def to_(self):
        """Return the data entities defined the 'to' depth intervals."""
        if (
            self.properties is None
            or len(self.properties) < 2
            or self._property_group_type != GroupTypeEnum.INTERVAL
        ):
            return None

        data = self.parent.get_data(  # pylint: disable=no-value-for-parameter
            self.properties[1]
        )

        if any(data) and isinstance(data[0], Data) and "to" in data[0].name.lower():
            return data[0]

        return None

    @property
    def parent(self) -> ConcatenatedObject:
        return self._parent

    @parent.setter
    def parent(self, parent):
        if self._parent is not None:
            raise AttributeError("Cannot change parent of a property group.")

        if not hasattr(parent, "add_children"):
            raise ValueError(
                "The 'parent' of a concatenated data must have an 'add_children' method."
            )

        parent.add_children([self])
        self._parent: ConcatenatedObject = parent

        parent.workspace.add_or_update_property_group(self)

    def _clear_data_list(
        self, data: str | Data | list[str | Data | UUID] | UUID
    ) -> list[str | Data | UUID]:
        """
        Clear the data list of any data that is a depth or from/to data.

        :param data: List of data to clear.

        :return: List of data with depth and from/to data removed.
        """
        if not isinstance(data, (list, tuple)):
            data = [data]

        # avoid suppressing depth and from-to directly
        avoid = []
        if self.depth_ is not None:
            avoid.extend([self.depth_, self.depth_.uid, self.depth_.name])
        if self.from_ is not None:
            avoid.extend([self.from_, self.from_.uid, self.from_.name])
        if self.to_ is not None:
            avoid.extend([self.to_, self.to_.uid, self.to_.name])

        return [d for d in data if d not in avoid]

    def _remove_depth_from_to(self):
        """
        Remove the depth or from/to data from the property group
            if they are the only properties left.
        """
        if (
            self._properties is not None
            and len(self._properties) == 1
            and self.depth_ is not None
        ):
            self.depth_.allow_delete = True
            self._properties.remove(self.depth_.uid)
            self.parent.remove_children([self.depth_, self])

        elif (
            self._properties is not None
            and len(self._properties) == 2
            and self.from_ is not None
            and self.to_ is not None
        ):
            self.to_.allow_delete = True
            self._properties.remove(self.to_.uid)
            self.parent.remove_children([self.to_])

            self.from_.allow_delete = True
            self._properties.remove(self.from_.uid)
            self.parent.remove_children([self.from_, self])

    def remove_properties(self, data: str | Data | list[str | Data | UUID] | UUID):
        """
        Remove data from the properties.

        The property group is removed if only the depth or from/to data are left.

        :param data: Data to remove.
        """
        data = self._clear_data_list(data)
        if data:
            super().remove_properties(data)

        self._remove_depth_from_to()



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-arguments, too-many-instance-attributes

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Literal, get_args
from uuid import UUID

import numpy as np

from ..shared import EntityType, utils
from .color_map import ColorMap
from .primitive_type_enum import PrimitiveTypeEnum
from .reference_value_map import BOOLEAN_VALUE_MAP, ReferenceValueMap


if TYPE_CHECKING:  # pragma: no cover
    from ..objects import ObjectBase
    from ..workspace import Workspace
    from .data import Data
    from .referenced_data import ReferencedData

ColorMapping = Literal[
    "linear",
    "equal_area",
    "logarithmic",
    "cdf",
    "cumulative_distribution_function",
    "missing",
]


class DataType(EntityType):
    """
    DataType class.

    Controls all the attributes of the data type for displays in Geoscience ANALYST.

    :param workspace: An active Workspace.
    :param primitive_type: The primitive type of the data.
    :param color_map: The colormap used for plotting.
    :param duplicate_on_copy: Force a copy on copy of the data entity.
    :param duplicate_type_on_copy: Force a copy on copy of the data entity.
    :param hidden: If the data are hidden or not.
    :param mapping: The type of color stretching to plot the colormap.
    :param number_of_bins: The number of bins used by the histogram.
    :param precision: The decimals precision of the data to display.
    :param scale: The type of scale of the data.
    :param scientific_notation: If the data should be displayed in scientific notation.
    :param transparent_no_data: If the no data values are displayed as transparent or not.
    :param units: The type of the units of the data.
    :param kwargs: Additional keyword arguments to set as attributes
        (see :obj:`...shared.entity_type.EntityType`).
    """

    _attribute_map = EntityType._attribute_map.copy()
    _attribute_map.update(
        {
            "Hidden": "hidden",
            "Mapping": "mapping",
            "Number of bins": "number_of_bins",
            "Precision": "precision",
            "Primitive type": "primitive_type",
            "Transparent no data": "transparent_no_data",
            "Scale": "scale",
            "Scientific notation": "scientific_notation",
        }
    )

    def __init__(
        self,
        workspace: Workspace,
        *,
        primitive_type: (
            type[Data] | PrimitiveTypeEnum | str
        ) = PrimitiveTypeEnum.INVALID,
        color_map: ColorMap | None = None,
        duplicate_on_copy: bool = False,
        duplicate_type_on_copy: bool = False,
        hidden: bool = False,
        mapping: ColorMapping = "equal_area",
        number_of_bins: int | None = None,
        precision: int = 2,
        scale: str | None = None,
        scientific_notation: bool = False,
        transparent_no_data: bool = True,
        units: str | None = None,
        **kwargs,
    ):
        super().__init__(workspace, **kwargs)

        self.color_map = color_map
        self.duplicate_on_copy = duplicate_on_copy
        self.duplicate_type_on_copy = duplicate_type_on_copy
        self.hidden = hidden
        self.mapping = mapping
        self.number_of_bins = number_of_bins
        self.precision = precision
        self.primitive_type = self.validate_primitive_type(primitive_type)
        self.scale = scale
        self.scientific_notation = scientific_notation
        self.transparent_no_data = transparent_no_data
        self.units = units

    @property
    def color_map(self) -> ColorMap | None:
        r"""
        The Colormap used for plotting

        The colormap can be set from a dictionary of sorted values with
        corresponding RGBA color.
        Or from a numpy array containing the RGBA values.

        .. code-block:: python

            color_map = {
                val_1: [r_1, g_1, b_1, a_1],
                ...,
                val_i: [r_i, g_i, b_i, a_i]
            }

        It can be set to None if non-existing.
        """
        return self._color_map

    @color_map.setter
    def color_map(self, color_map: ColorMap | dict | np.ndarray | None):
        if isinstance(color_map, dict):
            color_map = ColorMap(**color_map)

        elif isinstance(color_map, np.ndarray):
            color_map = ColorMap(values=color_map)

        if not isinstance(color_map, (ColorMap, type(None))):
            raise TypeError(
                f"Attribute 'color_map' must be of type {ColorMap},"
                f"numpy.ndarray or dict with 'values'."
            )

        if isinstance(color_map, ColorMap):
            color_map.parent = self

        self._color_map: ColorMap | None = color_map

        self.workspace.update_attribute(self, "color_map")

    @property
    def duplicate_on_copy(self) -> bool:
        """
        If the data type should be duplicated on copy.
        """
        return self._duplicate_on_copy

    @duplicate_on_copy.setter
    def duplicate_on_copy(self, value: bool):
        if not isinstance(value, bool) and value not in [1, 0]:
            raise TypeError(
                f"Attribute 'duplicate_on_copy' must be a bool, not {type(value)}"
            )

        self._duplicate_on_copy = bool(value)
        if self.on_file:
            self.workspace.update_attribute(self, "attributes")

    @property
    def duplicate_type_on_copy(self) -> bool:
        """
        If the data type should be duplicated on copy.
        """
        return self._duplicate_type_on_copy

    @duplicate_type_on_copy.setter
    def duplicate_type_on_copy(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(
                f"Attribute 'duplicate_type_on copy' must be a bool, not {type(value)}"
            )

        self._duplicate_type_on_copy = bool(value)
        self.workspace.update_attribute(self, "attributes")

    @classmethod
    def find_or_create_type(
        cls,
        workspace: Workspace,
        primitive_type: PrimitiveTypeEnum | str,
        dynamic_implementation_id: str | UUID | None = None,
        uid: UUID | None = None,
        **kwargs,
    ) -> DataType:
        """
        Get the data type for geometric data.

        :param workspace: An active Workspace class
        :param primitive_type: The primitive type of the data.
        :param uid: The unique identifier of the entity type.
        :param kwargs: The attributes of the entity type.

        :return: EntityType
        """
        if uid is not None:
            entity_type = DataType.find(workspace, uid)

            if entity_type is not None:
                return entity_type

        primitive_type = cls.validate_primitive_type(primitive_type)

        if primitive_type == PrimitiveTypeEnum.BOOLEAN:
            return ReferencedBooleanType(
                workspace, primitive_type=primitive_type, uid=uid, **kwargs
            )

        if (
            primitive_type == PrimitiveTypeEnum.GEOMETRIC
            and dynamic_implementation_id is not None
        ):
            data_type = DYNAMIC_CLASS_IDS.get(
                utils.str2uuid(dynamic_implementation_id), DataType
            )

            return data_type(
                workspace, primitive_type=primitive_type, uid=uid, **kwargs
            )

        if primitive_type == PrimitiveTypeEnum.REFERENCED:
            return ReferencedValueMapType(
                workspace, primitive_type=primitive_type, uid=uid, **kwargs
            )

        return cls(workspace, primitive_type=primitive_type, uid=uid, **kwargs)

    @property
    def hidden(self) -> bool:
        """
        If the data are hidden or not.
        """
        return self._hidden

    @hidden.setter
    def hidden(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(f"Attribute 'hidden' must be a bool, not {type(value)}")

        self._hidden: bool = bool(value)

        self.workspace.update_attribute(self, "attributes")

    @property
    def mapping(self) -> str:
        """
        The type of color stretching to plot the colormap.
        It chan be one of the following:
        'linear', 'equal_area', 'logarithmic', 'cdf', 'missing'
        """
        return self._mapping

    @mapping.setter
    def mapping(self, value: ColorMapping):
        if value not in get_args(ColorMapping):
            raise ValueError(
                f"Attribute 'mapping' should be one of {get_args(ColorMapping)}. "
                f"Value '{value}' was provided."
            )
        self._mapping: str = value

        self.workspace.update_attribute(self, "attributes")

    @property
    def number_of_bins(self) -> int | None:
        """
        The number of bins used by the histogram.
        It can be None if no histogram is used.
        """
        return self._number_of_bins

    @number_of_bins.setter
    def number_of_bins(self, n_bins: int | None):
        if n_bins is None:
            pass
        elif not isinstance(n_bins, (int, np.integer)) or n_bins < 1:
            raise ValueError(
                "Attribute 'number_of_bins' should be an integer greater than 0 "
                f"or None, not {n_bins}"
            )

        self._number_of_bins: int | None = n_bins

        self.workspace.update_attribute(self, "attributes")

    @property
    def precision(self) -> int:
        """
        The decimals precision of the data to display.
        """
        return self._precision

    @precision.setter
    def precision(self, value: int):
        if (
            not isinstance(value, (int, float, np.integer, np.floating))
            or (isinstance(value, (float, np.floating)) and not value.is_integer())
            or value < 0
        ):
            raise TypeError(
                f"Attribute 'precision' must be an integer greater than 0, not {value}"
            )

        self._precision = int(value)

        self.workspace.update_attribute(self, "attributes")

    @property
    def primitive_type(self) -> PrimitiveTypeEnum:
        """
        The primitive type of the data.
        """
        return self._primitive_type

    @primitive_type.setter
    def primitive_type(self, value: PrimitiveTypeEnum):
        if not isinstance(value, PrimitiveTypeEnum):
            raise ValueError(
                "Attribute 'primitive_type' value must be of type "
                f"{PrimitiveTypeEnum}, find {type(value)}"
            )

        self._primitive_type = value

    @property
    def transparent_no_data(self) -> bool:
        """
        If the no data values are displayed as transparent or not.
        """
        return self._transparent_no_data

    @transparent_no_data.setter
    def transparent_no_data(self, value: bool):
        if not isinstance(value, bool) and value != 1 and value != 0:
            raise TypeError(
                f"Attribute 'transparent_no_data' must be a bool, not {type(value)}"
            )
        self._transparent_no_data = bool(value)

        self.workspace.update_attribute(self, "attributes")

    @property
    def units(self) -> str | None:
        """
        The type of the units of the data.
        """
        return self._units

    @units.setter
    def units(self, unit: str | None):
        if not isinstance(unit, (str, type(None))):
            raise TypeError(f"Attribute 'units' must be a string, not {type(unit)}")
        self._units = unit

        self.workspace.update_attribute(self, "attributes")

    @staticmethod
    def primitive_type_from_values(values: np.ndarray | None) -> PrimitiveTypeEnum:
        """
        Validate the primitive type of the data.

        :param values: The values to validate.

        :return: The equivalent primitive type of the data.
        """
        if values is None or (
            isinstance(values, np.ndarray) and np.issubdtype(values.dtype, np.floating)
        ):
            primitive_type = PrimitiveTypeEnum.FLOAT

        elif isinstance(values, np.ndarray) and (
            np.issubdtype(values.dtype, np.integer)
        ):
            primitive_type = PrimitiveTypeEnum.INTEGER
        elif isinstance(values, str) or (
            isinstance(values, np.ndarray) and values.dtype.kind in ["U", "S"]
        ):
            primitive_type = PrimitiveTypeEnum.TEXT
        elif isinstance(values, np.ndarray) and (values.dtype == bool):
            primitive_type = PrimitiveTypeEnum.BOOLEAN
        else:
            raise NotImplementedError(
                "Only add_data values of type FLOAT, INTEGER,"
                "BOOLEAN and TEXT have been implemented"
            )
        return primitive_type

    @property
    def scale(self) -> str | None:
        """
        The type of scale of the data.
        """
        return self._scale

    @scale.setter
    def scale(self, value: str | None):
        if value not in ["Linear", "Log", None]:
            raise ValueError(
                f"Attribute 'scale' must be one of 'Linear', 'Log', NoneType, not {value}"
            )
        self._scale = value
        self.workspace.update_attribute(self, "attributes")

    @property
    def scientific_notation(self) -> bool:
        """
        If the data should be displayed in scientific notation.
        """
        return self._scientific_notation

    @scientific_notation.setter
    def scientific_notation(self, value: bool):
        if value not in [True, False, 1, 0]:
            raise TypeError(
                f"Attribute 'scientific_notation' must be a bool, not {type(value)}"
            )

        self._scientific_notation = bool(value)
        self.workspace.update_attribute(self, "attributes")

    @staticmethod
    def validate_primitive_type(
        primitive_type: PrimitiveTypeEnum | str | type[Data],
    ) -> PrimitiveTypeEnum:
        """
        Validate the primitive type of the data.

        :param primitive_type: Some reference to the primitive type of the data.

        :return: A known primitive type.
        """
        if isinstance(primitive_type, str):
            primitive_type = getattr(
                PrimitiveTypeEnum, utils.INV_KEY_MAP.get(primitive_type, primitive_type)
            )

        if isinstance(primitive_type, type) and hasattr(
            primitive_type, "primitive_type"
        ):
            primitive_type = primitive_type.primitive_type()

        if not isinstance(primitive_type, PrimitiveTypeEnum):
            raise ValueError(
                f"Attribute 'primitive_type' should be one of {PrimitiveTypeEnum.__members__}"
            )
        return primitive_type


class ReferenceDataType(DataType):
    """
    DataType class.

    Controls all the attributes of reference data.

    :param value_map: Reference value to map index with description.
    """

    def __init__(
        self,
        workspace: Workspace,
        value_map: dict[int, str] | np.ndarray | tuple | ReferenceValueMap = (
            (0, "Unknown"),
        ),
        **kwargs,
    ):
        super().__init__(workspace, **kwargs)
        self.value_map = self.validate_value_map(value_map)

    @staticmethod
    @abstractmethod
    def validate_keys(value_map: ReferenceValueMap) -> ReferenceValueMap:
        """
        Validate the keys of the value map.
        """

    def validate_value_map(
        self,
        value_map: dict[int, str] | np.ndarray | tuple | ReferenceValueMap,
    ) -> ReferenceValueMap | None:
        """
        Validate the attribute of ReferencedDataType
        """

        if value_map is None:
            return None

        if isinstance(value_map, dict | np.ndarray | tuple):
            value_map = ReferenceValueMap(value_map)

        if not isinstance(value_map, ReferenceValueMap):
            raise TypeError(
                "Attribute 'value_map' must be provided as a dict, tuple[dict], "
                f"numpy.ndarray or {ReferenceValueMap}."
            )

        self.validate_keys(value_map)

        return value_map

    @property
    def value_map(self) -> ReferenceValueMap | None:
        r"""
        Reference value map for to map index with description.

        The value_map can be set from a dictionary of sorted values int
        values with text description.

        .. code-block:: python

            value_map = {
                val_1: str_1,
                ...,
                val_i: str_i
            }

        """
        return self._value_map

    @value_map.setter
    def value_map(
        self, value_map: dict[int, str] | np.ndarray | tuple | ReferenceValueMap
    ):
        self._value_map = self.validate_value_map(value_map)

        if self.on_file:
            self.workspace.update_attribute(self, "value_map")


class ReferencedValueMapType(ReferenceDataType):
    """
    Data container for referenced value map.
    """

    _TYPE_UID = UUID(fields=(0x2D5D6C1E, 0x4D8C, 0x4F3A, 0x9B, 0x3F, 0x2E5A0D8E1C1F))

    def __init__(
        self,
        workspace: Workspace,
        **kwargs,
    ):
        super().__init__(workspace, **kwargs)

    @staticmethod
    def validate_keys(value_map: ReferenceValueMap):
        """
        Validate the keys of the value map.
        """
        if 0 not in value_map.map["Key"]:
            value_map.map.resize(len(value_map) + 1, refcheck=False)
            value_map.map[-1] = (0, b"Unknown")

        if dict(value_map.map)[0] not in ["Unknown", b"Unknown"]:
            raise ValueError("Value for key 0 must be b'Unknown'")


class ReferencedBooleanType(ReferenceDataType):
    """
    Data container for referenced boolean data.
    """

    def __init__(
        self,
        workspace: Workspace,
        value_map: (
            dict[int, str] | np.ndarray | tuple | ReferenceValueMap
        ) = BOOLEAN_VALUE_MAP,
        **kwargs,
    ):
        super().__init__(workspace, value_map=value_map, **kwargs)

    @staticmethod
    def validate_keys(value_map: ReferenceValueMap):
        """
        Validate the keys of the value map.
        """
        if not np.all(value_map.map == BOOLEAN_VALUE_MAP):
            raise ValueError("Boolean value map must be (0: 'False', 1: 'True'")


class GeometricDynamicDataType(DataType, ABC):
    """
    Data container for dynamic geometric data.
    """

    _attribute_map = DataType._attribute_map.copy()
    _attribute_map.update(
        {
            "Dynamic implementation ID": "dynamic_implementation_id",
        }
    )
    _TYPE_UID: UUID | None
    _DYNAMIC_IMPLEMENTATION_ID: UUID

    def __init__(
        self,
        workspace: Workspace,
        uid: UUID | None = None,
        **kwargs,
    ):
        if uid is None:
            uid = self._TYPE_UID

        super().__init__(workspace, uid=uid, **kwargs)

    @classmethod
    def default_type_uid(cls) -> UUID | None:
        """
        Default uuid for the entity type.
        """
        return cls._TYPE_UID

    @property
    def dynamic_implementation_id(self) -> UUID:
        """
        The dynamic implementation id.
        """
        return self._DYNAMIC_IMPLEMENTATION_ID


class GeometricDataValueMapType(ReferenceDataType, GeometricDynamicDataType):
    """
    Data container for value map
    """

    _DYNAMIC_IMPLEMENTATION_ID = UUID("{4b6ecb37-0623-4ea0-95f1-4873008890a8}")
    _TYPE_UID = None

    def __init__(
        self,
        workspace: Workspace,
        *,
        value_map: dict[int, str] | tuple | ReferenceValueMap | None = None,
        parent: ObjectBase | None = None,
        description: str = "Dynamic referenced data",
        primitive_type: PrimitiveTypeEnum | str = PrimitiveTypeEnum.GEOMETRIC,
        **kwargs,
    ):
        self._referenced_data = None

        super().__init__(
            workspace,
            value_map=value_map,
            description=description,
            primitive_type=primitive_type,
            **kwargs,
        )
        self._parent = parent

    def get_parent_reference(self, parent: ObjectBase):
        """
        Recover the parent ReferencedData by name.
        """
        ref_data_name = self.name.rsplit(":")[0]

        ref_data = []
        for child in parent.children:
            if (
                isinstance(child.entity_type, ReferencedValueMapType)
                and child.entity_type.name == ref_data_name
            ):
                ref_data.append(child)

        if len(ref_data) == 0:
            raise ValueError(f"Parent data '{ref_data_name}' not found.")

        return ref_data[0]

    @property
    def referenced_data(self) -> ReferencedData | None:
        """
        Reference data type holding the value map.
        """
        if self._referenced_data is None and self._parent is not None:
            self._referenced_data = self.get_parent_reference(self._parent)

        return self._referenced_data

    @staticmethod
    def validate_keys(value_map: ReferenceValueMap):
        """
        Validate the keys of the value map.
        """

    def validate_value_map(
        self,
        value_map: dict[int, str] | np.ndarray | tuple | ReferenceValueMap,
    ) -> ReferenceValueMap | None:
        """
        Validate the attribute of ReferencedDataType
        """

        if value_map is None:
            return None

        if isinstance(value_map, dict | np.ndarray | tuple):
            value_map = ReferenceValueMap(value_map, name=self.name.rsplit(": ")[1])

        if not isinstance(value_map, ReferenceValueMap):
            raise TypeError(
                "Attribute 'value_map' must be provided as a dict, tuple[dict] "
                f"or {ReferenceValueMap}."
            )

        return value_map

    @property
    def value_map(self) -> ReferenceValueMap | None:
        r"""
        Reference value to map index with description.

        The value_map can be set from a dictionary of sorted integer
        values with text description.

        .. code-block:: python

            value_map = {
                val_1: str_1,
                ...,
                val_i: str_i
            }

        """
        if self._value_map is None and self.referenced_data is not None:
            if (
                self.referenced_data.data_maps is None
                or self.referenced_data.metadata is None
            ):
                raise ValueError("Referenced data has no data maps.")

            value_map = None
            for count, name in enumerate(self.referenced_data.metadata):
                if name == self.name.rsplit(": ")[1]:
                    value_map = self.workspace.fetch_array_attribute(
                        self.referenced_data.entity_type, f"Value map {count + 1}"
                    )

            if value_map is not None:
                self._value_map = self.validate_value_map(
                    value_map.astype(ReferenceValueMap.MAP_DTYPE)
                )

        return self._value_map

    @value_map.setter
    def value_map(self, value_map: dict | tuple | ReferenceValueMap | None):
        self._value_map = self.validate_value_map(value_map)

        if self.on_file and self.referenced_data is not None:
            self.workspace.update_attribute(self.referenced_data, "data_map")


class GeometricDataXType(GeometricDynamicDataType):
    """
    Data container for X values
    """

    _DYNAMIC_IMPLEMENTATION_ID = UUID("{2dbf303e-05d6-44ba-9692-39474e88d516}")
    _TYPE_UID = UUID(fields=(0xE9E6B408, 0x4109, 0x4E42, 0xB6, 0xA8, 0x685C37A802EE))


class GeometricDataYType(GeometricDynamicDataType):
    """
    Data container for Y values
    """

    _DYNAMIC_IMPLEMENTATION_ID = UUID("{d56406dc-5eeb-418d-add4-a1282a6ef668}")
    _TYPE_UID = UUID(fields=(0xF55B07BD, 0xD8A0, 0x4DFF, 0xBA, 0xE5, 0xC975D490D71C))


class GeometricDataZType(GeometricDynamicDataType):
    """
    Data container for Z values
    """

    _DYNAMIC_IMPLEMENTATION_ID = UUID("{9dacdc3b-6878-408d-93ae-e9a95e640f0c}")
    _TYPE_UID = UUID(fields=(0xDBAFB885, 0x1531, 0x410C, 0xB1, 0x8E, 0x6AC9A40B4466))


DYNAMIC_CLASS_IDS = {
    cls._DYNAMIC_IMPLEMENTATION_ID: cls  # pylint: disable=protected-access
    for cls in GeometricDynamicDataType.__subclasses__()
}



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

from abc import ABC, abstractmethod
from warnings import warn

import numpy as np

from ...data import NumericData
from ...objects import ObjectBase
from ...workspace import Workspace


class BaseMerger(ABC):
    _type: type = ObjectBase

    @classmethod
    def merge_data(
        cls,
        out_entity,
        input_entities: list,
    ):
        """
        Merge the data respecting the entity type, the values, and the association.
        :param out_entity: the output entity to add the data to.
        :param input_entities: the list of objects to merge the data from.
        :return: a dictionary of data to add to the merged object.
        """
        data_dict: dict[tuple, NumericData] = {}
        data_count = {"VERTEX": 0, "CELL": 0}

        for input_entity in input_entities:
            for ind, data in enumerate(input_entity.children):
                if (
                    not isinstance(data, NumericData)
                    or data.association is None
                    or data.n_values is None
                    or data.values is None
                ):
                    continue

                association = data.association.name
                label = (data.name, data.entity_type.name, association)
                start, end = (
                    data_count[association],
                    data_count[association] + data.n_values,
                )

                # Increment label in case of duplicate match
                if label in data_dict:
                    values = data_dict[label].values
                    if isinstance(values, np.ndarray) and (
                        ~np.all(
                            np.logical_or(
                                values[start:end] == data.nan_value,
                                np.isnan(values[start:end]),
                            )
                        )
                    ):
                        label = (
                            data.name + f"({ind})",
                            data.entity_type.name,
                            association,
                        )
                        warn(
                            f"Multiple data '{data.name}' with entity_type "
                            f"name '{data.entity_type.name}' "
                            f"were found on object '{input_entity.name}'. "
                            f"The merging operation is ambiguous. "
                            f"Please validate or rename the data."
                        )

                if label not in data_dict:
                    shape = (
                        out_entity.n_vertices
                        if association == "VERTEX"
                        else out_entity.n_cells
                    )

                    data_dict[label] = out_entity.add_data(
                        {
                            data.name: {
                                "values": np.ones(shape) * data.nan_value,
                                "association": association,
                                "entity_type": data.entity_type,
                            }
                        }
                    )

                values = data_dict[label].values

                if isinstance(values, np.ndarray):
                    values[start:end] = data.values
                    data_dict[label].values = values

            data_count["VERTEX"] += getattr(input_entity, "n_vertices", 0) or 0
            data_count["CELL"] += getattr(input_entity, "n_cells", 0) or 0

    @classmethod
    def merge_objects(
        cls,
        workspace: Workspace,
        input_entities: list[ObjectBase],
        add_data: bool = True,
        **kwargs,
    ):
        """
        Merge a list of :obj:geoh5py.objects.ObjectBase of the same type.

        :param workspace: the workspace to use.
        :param input_entities: the list of objects to merge.
        :param add_data: if True, the data will be merged.
        :param kwargs: the kwargs to create the new object.
        :return: an object of type input_entities[0] containing the merged vertices.
        """
        cls.validate_objects(input_entities)
        kwargs.pop("workspace", None)

        output = cls.create_object(workspace, input_entities, **kwargs)

        # add the data
        if add_data:
            cls.merge_data(output, input_entities)

        return output

    @classmethod
    @abstractmethod
    def create_object(cls, workspace: Workspace, input_entities, **kwargs):
        """
        Create an object with the structure of all the input entities.
        :param workspace: The workspace to use to create the object.
        :param input_entities: all the input entities to merge.
        :param kwargs: the kwargs to create the new object.
        :return: the new object.
        """
        raise NotImplementedError("BaseMerger cannot be use, use a subclass.")

    @classmethod
    def validate_objects(cls, input_entities: list[ObjectBase]):
        """
        Validate the input entities types and raises error if incompatible.

        :param input_entities: a list of :obj:geoh5py.objects.ObjectBase objects.
        """
        # assert input entities is a list of len superior to 1
        if not isinstance(input_entities, list):
            raise TypeError("The input entities must be a list of geoh5py objects.")

        # assert input entities is a list of len superior to 1
        if len(input_entities) < 2:
            raise ValueError("Need more than one object to merge.")

        # assert input entities are of the same type
        if not all(
            type(input_entity) is type(input_entities[0])
            for input_entity in input_entities
        ):
            raise TypeError("All objects must be of the same type.")

        # assert input entities are of the same type
        cls.validate_type(input_entities[0])

        # verify if the all input entities have vertices
        for input_entity in input_entities:
            cls.validate_structure(input_entity)

    @classmethod
    @abstractmethod
    def validate_structure(cls, input_entity):
        """
        Validate the input entity structure and raises error if incompatible.
        :param input_entity: the input entity to validate.
        """
        raise NotImplementedError("BaseMerger cannot be use, use a subclass.")

    @classmethod
    def validate_type(cls, input_entity):
        if type(input_entity) is not cls._type:  # pylint: disable=unidiomatic-typecheck
            raise TypeError(
                f"The input entities must be a list of {cls._type.__name__}."
            )



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-many-lines, too-many-public-methods, too-many-arguments, too-many-locals

from __future__ import annotations

import inspect
import subprocess
import tempfile
import uuid
import warnings
import weakref
from contextlib import AbstractContextManager, contextmanager
from gc import collect
from getpass import getuser
from io import BytesIO
from pathlib import Path
from shutil import copy, move
from subprocess import CalledProcessError
from typing import Any, ClassVar, cast
from weakref import ReferenceType

import h5py
import numpy as np

from .. import data, groups, objects
from ..data import CommentsData, Data, DataType, PrimitiveTypeEnum
from ..data.text_data import TextData
from ..data.visual_parameters import VisualParameters
from ..groups import (
    CustomGroup,
    DrillholeGroup,
    Group,
    IntegratorDrillholeGroup,
    PropertyGroup,
    RootGroup,
)
from ..io import H5Reader, H5Writer
from ..io.utils import str_from_subtype, str_from_type
from ..objects import Drillhole, ObjectBase
from ..shared import weakref_utils
from ..shared.concatenation import (
    Concatenated,
    ConcatenatedData,
    ConcatenatedDrillhole,
    ConcatenatedObject,
    ConcatenatedPropertyGroup,
    Concatenator,
)
from ..shared.entity import Entity
from ..shared.entity_type import EntityType
from ..shared.exceptions import Geoh5FileClosedError
from ..shared.utils import (
    as_str_if_utf8_bytes,
    clear_array_attributes,
    get_attributes,
    str2uuid,
)


# pylint: disable=too-many-instance-attributes
class Workspace(AbstractContextManager):
    """
    The Workspace class manages all Entities created or imported from the
    *geoh5* structure.

    The basic requirements needed to create a Workspace are:

    :param h5file: Path to the *geoh5* file or :obj:`oi.BytesIO` representation
        of a geoh5 structure.
    :param contributors: List of contributors to the project.
    :param distance_unit: Distance unit used in the project.
    :param ga_version: Version of the *geoh5* file format.
    :param mode: Mode in which the *geoh5* file is opened.
    :param name: Name of the project.
    :param repack: Repack the *geoh5* file after closing.
    :param version: Version of the project.
    """

    _active_ref: ClassVar[ReferenceType[Workspace]] | type(None) = type(None)  # type: ignore
    _attribute_map = {
        "Contributors": "contributors",
        "Distance unit": "distance_unit",
        "GA Version": "ga_version",
        "Version": "version",
    }

    def __init__(
        self,
        h5file: str | Path | BytesIO | None = None,
        *,
        contributors: tuple[str] = (getuser(),),
        distance_unit: str = "meter",
        ga_version: str = "1",
        mode="r+",
        name: str = "GEOSCIENCE",
        repack: bool = False,
        version: float = 2.1,
    ):
        self._root: RootGroup
        self._data: dict[uuid.UUID, ReferenceType[data.Data]] = {}
        self._distance_unit: str = distance_unit
        self._contributors: np.ndarray = np.asarray(
            contributors, dtype=h5py.special_dtype(vlen=str)
        )
        self._ga_version: str = ga_version
        self._geoh5: h5py.File | bool = False
        self._groups: dict[uuid.UUID, ReferenceType[Group]] = {}
        self._property_groups: dict[uuid.UUID, ReferenceType[PropertyGroup]] = {}
        self._h5file: str | Path | BytesIO | None = None
        self._mode: str = mode
        self._name: str = name
        self._objects: dict[uuid.UUID, ReferenceType[ObjectBase]] = {}
        self._repack: bool = repack
        self._types: dict[uuid.UUID, ReferenceType[EntityType]] = {}
        self._version: float = version

        self.h5file = h5file
        self.open()

    def activate(self):
        """Makes this workspace the active one.

        In case the workspace gets deleted, Workspace.active() safely returns None.
        """
        if Workspace._active_ref() is not self:
            Workspace._active_ref = weakref.ref(self)

    @staticmethod
    def active() -> Workspace:
        """Get the active workspace."""
        active_one = Workspace._active_ref()
        if active_one is None:
            raise RuntimeError("No active workspace.")

        # so that type check does not complain of possible returned None
        return cast(Workspace, active_one)

    def _all_data(self) -> list[data.Data]:
        """Get all active Data entities registered in the workspace."""
        self.remove_none_referents(self._data, "Data")
        return [cast("Data", v()) for v in self._data.values()]

    def _all_groups(self) -> list[groups.Group]:
        """Get all active Group entities registered in the workspace."""
        self.remove_none_referents(self._groups, "Groups")
        return [cast("Group", v()) for v in self._groups.values()]

    def _all_property_groups(self) -> list[PropertyGroup]:
        """Get all active PropertyGroup entities registered in the workspace."""
        self.remove_none_referents(self._property_groups, "PropertyGroups")
        return [cast("PropertyGroup", v()) for v in self._property_groups.values()]

    def _all_objects(self) -> list[objects.ObjectBase]:
        """Get all active Object entities registered in the workspace."""
        self.remove_none_referents(self._objects, "Objects")
        return [cast("ObjectBase", v()) for v in self._objects.values()]

    def _all_types(self) -> list[EntityType]:
        """Get all active entity types registered in the workspace."""
        self.remove_none_referents(self._types, "Types")
        return [cast("EntityType", v()) for v in self._types.values()]

    @property
    def attribute_map(self) -> dict:
        """
        Mapping between names used in the geoh5 database.
        """
        return self._attribute_map

    def close(self):
        """
        Close the file and clear properties for future open.
        """
        if not self._geoh5:
            return

        if self.geoh5.mode in ["r+", "a"]:
            for entity in self.groups:
                if isinstance(entity, Concatenator) and self.repack:
                    self.update_attribute(entity, "concatenated_attributes")

            self._io_call(H5Writer.save_entity, self.root, add_children=True, mode="r+")

        self.geoh5.close()

        if (
            self.repack
            and not isinstance(self._h5file, BytesIO)
            and self._h5file is not None
        ):
            temp_file = Path(tempfile.gettempdir()) / Path(self._h5file).name
            try:
                subprocess.run(
                    f'h5repack --native "{self._h5file}" "{temp_file}"',
                    check=True,
                    shell=True,
                    stdout=subprocess.DEVNULL,
                )
                Path(self._h5file).unlink()
                move(temp_file, self._h5file, copy)
            except CalledProcessError:
                pass

            self.repack = False

    @property
    def contributors(self) -> np.ndarray:
        """
        :obj:`numpy.array` of :obj:`str` List of contributors name.
        """
        return self._contributors

    @contributors.setter
    def contributors(self, value: list[str]):
        self._contributors = np.asarray(value, dtype=h5py.special_dtype(vlen=str))

    def copy_to_parent(
        self,
        entity,
        parent,
        omit_list: tuple = (),
        clear_cache: bool = False,
        **kwargs,
    ):
        """
        Copy an entity to a different parent with copies of children.

        :param entity: Entity to be copied.
        :param parent: Target parent to copy the entity under.
        :param omit_list: List of property names to omit on copy
        :param clear_cache: Clear array attributes after copy.
        :param kwargs: Additional keyword arguments passed to the copy constructor.

        :return: The Entity registered to the workspace.
        """
        entity_kwargs = get_attributes(
            entity,
            omit_list=[
                "_uid",
                "_entity_type",
                "_on_file",
                "_centroids",
                "_parts",
                "_extent",
                "_visual_parameters",
            ]
            + list(omit_list),
            attributes={"uid": None, "parent": None},
        )

        if entity_kwargs is None:
            return None

        entity_type_kwargs = get_attributes(
            entity.entity_type,
            omit_list=["_workspace", "_on_file"] + list(omit_list),
        )

        # overwrite kwargs
        entity_kwargs.update(
            (k, kwargs[k]) for k in entity_kwargs.keys() & kwargs.keys()
        )
        entity_type_kwargs.update(
            (k, kwargs[k]) for k in entity_type_kwargs.keys() & kwargs.keys()
        )

        if not isinstance(parent, (ObjectBase, Group, Workspace)):
            raise ValueError(
                "Input 'parent' should be of type (ObjectBase, Group, Workspace)"
            )

        if isinstance(parent, Workspace):
            parent = parent.root

        # Assign the same uid if possible
        if parent.workspace.get_entity(entity.uid)[0] is None:
            entity_kwargs["uid"] = entity.uid

        entity_kwargs["parent"] = parent

        entity_type = type(entity)
        if isinstance(entity, Data):
            entity_type = Data

        entity_kwargs.pop("property_groups", None)

        new_object = parent.workspace.create_entity(
            entity_type, entity=entity_kwargs, entity_type=entity_type_kwargs
        )

        if clear_cache:
            clear_array_attributes(entity)
            clear_array_attributes(new_object)

        return new_object

    @classmethod
    def copy_property_groups(
        cls, entity: ObjectBase, property_groups: list[PropertyGroup], data_map: dict
    ):
        """
        Copy property groups to a new entity.
        Keep the same uid if it's not present on the new workspace.

        :param entity: The entity associated to the property groups.
        :param property_groups: The property groups to copy.
        :param data_map: the data map to use for the property groups.
        """
        for prop_group in property_groups:
            properties = None
            if prop_group.properties is not None:
                properties = [data_map[uid] for uid in prop_group.properties]

            # prepare the kwargs
            property_group_kwargs = {
                "association": prop_group.association,
                "name": prop_group.name,
                "property_group_type": prop_group.property_group_type,
                "properties": properties,
            }

            # Assign the same uid if possible
            if entity.workspace.find_property_group(prop_group.uid) is None:
                property_group_kwargs["uid"] = prop_group.uid

            entity.fetch_property_group(**property_group_kwargs)

    @classmethod
    def create(cls, path: str | Path, **kwargs) -> Workspace:
        """Create a named blank workspace and save to disk."""
        return cls(**kwargs).save_as(path)

    def create_from_concatenation(self, attributes):
        if "Name" in attributes:
            attributes["Name"] = attributes["Name"].replace("\u2044", "/")

        recovered_entity = None

        if "Object Type ID" in attributes:
            recovered_entity = self.create_entity(
                ObjectBase,
                save_on_creation=False,
                entity=attributes,
                entity_type={"uid": attributes.pop("Object Type ID")},
            )

        elif "Type ID" in attributes:
            recovered_entity = self.create_entity(
                Data,
                save_on_creation=False,
                entity=attributes,
                entity_type=self.fetch_type(
                    uuid.UUID(attributes.pop("Type ID")), "Data"
                ),
            )

        if recovered_entity is not None:
            recovered_entity.on_file = True
            recovered_entity.entity_type.on_file = True

        return recovered_entity

    def create_data(
        self,
        entity_class,
        entity: dict,
        entity_type: dict | EntityType,
    ) -> Data:
        """
        Create a new Data entity with attributes.

        :param entity_class: :obj:`~geoh5py.data.data.Data` class.
        :param entity: Properties of the entity.
        :param entity_type: Properties of the entity_type.

        :return: The newly created entity.
        """
        if isinstance(entity_type, dict):
            entity_type = DataType.find_or_create_type(
                self,
                entity_type.pop("primitive_type"),
                parent=entity["parent"],
                **entity_type,
            )
        elif not isinstance(entity_type, DataType):
            raise TypeError(
                f"Expected `entity_type` to be of type `dict` or `DataType`, "
                f"got {type(entity_type)}."
            )

        for name, member in inspect.getmembers(data):
            if (
                inspect.isclass(member)
                and issubclass(member, entity_class)
                and member is not entity_class
                and hasattr(member, "primitive_type")
                and inspect.ismethod(member.primitive_type)
                and entity_type.primitive_type is member.primitive_type()
            ):
                if member is CommentsData and not any(
                    isinstance(val, str) and val == "UserComments"
                    for val in entity.values()
                ):
                    continue

                if self.version > 1.0 and isinstance(
                    entity["parent"], ConcatenatedObject
                ):
                    member = type("Concatenated" + name, (ConcatenatedData, member), {})

                if member is TextData and any(
                    isinstance(val, str) and "Visual Parameters" == val
                    for val in entity.values()
                ):
                    member = VisualParameters

                created_entity = member(entity_type=entity_type, **entity)

                return created_entity

        raise TypeError(
            f"Data type {entity_class} not found in {entity_type.primitive_type}."
        )

    def create_entity(
        self,
        entity_class,
        *,
        compression: int = 5,
        entity: dict | None = None,
        entity_type: EntityType | dict | None = None,
        save_on_creation: bool = True,
    ):
        """
        Function to create and register a new entity and its entity_type.

        :param entity_class: Type of entity to be created
        :param compression: Compression level for data.
        :param entity: Attributes of the entity.
        :param entity_type: Attributes of the entity_type.
        :param save_on_creation: Save the entity to geoh5 immediately

        :return entity: Newly created entity registered to the workspace
        """

        entity = entity or {}
        entity_type = entity_type or {}

        if entity_class is not RootGroup and (
            "parent" not in entity or entity["parent"] is None
        ):
            entity["parent"] = self.root

        if entity_class is None or issubclass(entity_class, Data):
            created_data = self.create_data(Data, entity, entity_type)
            if save_on_creation and self.h5file is not None:
                self.save_entity(created_data, compression=compression)
            return created_data

        created_entity = self.create_object_or_group(entity_class, entity, entity_type)
        if save_on_creation and self.h5file is not None:
            self.save_entity(created_entity, compression=compression)
        return created_entity

    def add_or_update_property_group(
        self, property_group: PropertyGroup, remove: bool = False
    ):
        """
        Add or update a property group to the workspace.
        """
        if isinstance(property_group, ConcatenatedPropertyGroup):
            parent = property_group.parent
            parent.concatenator.update_attributes(parent, "property_groups")
        else:
            self._io_call(
                H5Writer.add_or_update_property_group,
                property_group,
                remove=remove,
                mode="r+",
            )

    def create_object_or_group(
        self, entity_class, entity: dict, entity_type: dict | EntityType
    ) -> Group | ObjectBase:
        """
        Create an object or a group with attributes.

        :param entity_class: :obj:`~geoh5py.objects.object_base.ObjectBase` or
            :obj:`~geoh5py.groups.group.Group` class.
        :param entity: Attributes of the entity.
        :param entity_type: Attributes of the entity_type.

        :return: A new Object or Group.
        """
        entity_type_uid = None

        if isinstance(entity_type, EntityType):
            entity_type = get_attributes(entity_type)

        for key, val in entity_type.items():
            if key.lower() in ["id", "uid"]:
                entity_type_uid = uuid.UUID(str(val))

        if entity_type_uid is None:
            if hasattr(entity_class, "default_type_uid"):
                entity_type_uid = entity_class.default_type_uid()
            else:
                entity_type_uid = uuid.uuid4()

        for name, member in inspect.getmembers(groups) + inspect.getmembers(objects):
            if (
                inspect.isclass(member)
                and issubclass(member, entity_class.__bases__)
                and member is not entity_class.__bases__
                and not member == CustomGroup
                and member.default_type_uid() == entity_type_uid
            ):
                if self.version > 1.0:
                    if member in (DrillholeGroup, IntegratorDrillholeGroup):
                        member = type("Concatenator" + name, (Concatenator, member), {})
                    elif member is Drillhole and isinstance(
                        entity.get("parent"),
                        (DrillholeGroup, IntegratorDrillholeGroup),
                    ):
                        member = ConcatenatedDrillhole

                entity_type = member.find_or_create_type(self, **entity_type)

                created_entity = member(entity_type=entity_type, **entity)

                return created_entity

        # Special case for CustomGroup without uuid
        if entity_class == Group:
            entity_type = groups.custom.CustomGroup.find_or_create_type(
                self, **entity_type
            )
            created_entity = groups.custom.CustomGroup(
                entity_type=entity_type, **entity
            )

            return created_entity

        raise TypeError(f"Entity class type {entity_class} not recognized.")

    def create_root(
        self, entity_attributes: dict | None = None, type_attributes: dict | None = None
    ) -> RootGroup:
        """
        Create a RootGroup entity.

        :param entity_attributes: Attributes of the entity.
        :param type_attributes: Attributes of the entity_type.

        :return: The newly created RootGroup entity.
        """
        type_attributes = type_attributes or {}
        group_type = RootGroup.find_or_create_type(self, **type_attributes)

        entity_attributes = entity_attributes or {}
        root = RootGroup(entity_type=group_type, **entity_attributes)

        return root

    @property
    def data(self) -> list[data.Data]:
        """Get all active Data entities registered in the workspace."""
        return self._all_data()

    def fetch_or_create_root(self):
        """
        Fetch the root group or create a new one if it does not exist.
        """
        attrs, type_attrs, _ = self._io_call(
            H5Reader.fetch_attributes, uuid.uuid4(), "root", mode="r"
        )
        self._root = self.create_root(
            entity_attributes=attrs, type_attributes=type_attrs
        )

        if attrs is not None:
            self._root.on_file = True
            self._root.entity_type.on_file = True
            self.fetch_children(self._root, recursively=True)

            return

        # Fetch all entities and build the family tree with RootGroup at the base
        for entity_type in ["group", "object"]:
            uuids = self._io_call(H5Reader.fetch_uuids, entity_type, mode="r")

            for uid in uuids:
                if isinstance(self.get_entity(uid)[0], Entity):
                    continue

                recovered_object = self.load_entity(uid, entity_type)

                if isinstance(recovered_object, (Group, ObjectBase)):
                    self.fetch_children(recovered_object, recursively=True)

    def remove_children(self, parent, children: list):
        """
        Remove a list of entities from a parent. The target entities remain
        present on file.
        """
        for child in children:
            if isinstance(child, PropertyGroup):
                self._io_call(
                    H5Writer.add_or_update_property_group, child, remove=True, mode="r+"
                )
            else:
                ref_type = str_from_type(child)
                self._io_call(
                    H5Writer.remove_child, child.uid, ref_type, parent, mode="r+"
                )

    def remove_entity(self, entity: Entity | PropertyGroup | EntityType, force=False):
        """
        Function to remove an entity and its children from the workspace.
        """
        if not getattr(entity, "allow_delete", True) and not force:
            raise UserWarning(
                f"The 'allow_delete' property of entity {entity} prevents it from "
                "being removed. Please revise."
            )

        if isinstance(entity, (Concatenated, ConcatenatedPropertyGroup)):
            entity.concatenator.remove_entity(entity)
            return

        if isinstance(entity, Entity | PropertyGroup):
            self.workspace.remove_recursively(entity)
            entity.parent.remove_children([entity])

        if not isinstance(entity, PropertyGroup):
            ref_type = str_from_type(entity)
            self._io_call(
                H5Writer.remove_entity,
                entity.uid,
                ref_type,
                mode="r+",
            )

        del entity
        collect()
        self.remove_none_referents(self._types, "Types")

    def remove_none_referents(
        self,
        referents: dict[uuid.UUID, ReferenceType],
        rtype: str,
    ):
        """
        Search and remove deleted entities
        """
        rem_list: list = []
        for key, value in referents.items():
            if value() is None:
                rem_list += [key]
                self._io_call(
                    H5Writer.remove_entity, key, rtype, parent=self, mode="r+"
                )

        for key in rem_list:
            del referents[key]

    def remove_recursively(self, entity: Entity | PropertyGroup):
        """Delete an entity and its children from the workspace and geoh5 recursively"""
        if hasattr(entity, "children"):
            for child in entity.children:
                self.remove_entity(child, force=True)

    def deactivate(self):
        """Deactivate this workspace if it was the active one, else does nothing."""
        if Workspace._active_ref() is self:
            Workspace._active_ref = type(None)

    @property
    def distance_unit(self) -> str:
        """
        :obj:`str` Distance unit used in the project.
        """
        return self._distance_unit

    @distance_unit.setter
    def distance_unit(self, value: str):
        self._distance_unit = value

    def fetch_array_attribute(
        self, entity: Entity | EntityType, key: str = "cells"
    ) -> np.ndarray:
        """
        Fetch attribute stored as structured array from the source geoh5.

        :param entity: Unique identifier of target entity.
        :param key: Field array name

        :return: Structured array.
        """
        if isinstance(entity, Concatenated):
            return entity.concatenator.fetch_values(entity, key)  # type: ignore

        if isinstance(entity, EntityType):
            entity_type = str_from_subtype(entity)
        else:
            entity_type = str_from_type(entity)

        return self._io_call(
            H5Reader.fetch_array_attribute,
            entity.uid,
            entity_type,
            key,
            mode="r",
        )

    def fetch_children(
        self,
        entity: Entity | PropertyGroup | None,
        recursively: bool = False,
    ) -> list:
        """
        Recover and register children entities from the geoh5.

        :param entity: Parental entity.
        :param recursively: Recover all children down the project tree.
        :return list: List of children entities.
        """
        if entity is None or isinstance(entity, ConcatenatedData):
            return []

        entity_type = str_from_type(entity)

        if isinstance(entity, RootGroup) and not entity.on_file:
            children_list = {child.uid: "" for child in entity.children}

        else:
            children_list = self._io_call(
                H5Reader.fetch_children, entity.uid, entity_type, mode="r"
            )

            if isinstance(entity, Concatenator):
                if any(entity.children):
                    return entity.children

                cat_children = entity.fetch_concatenated_objects()
                children_list.update(
                    {
                        str2uuid(as_str_if_utf8_bytes(uid)): attr
                        for uid, attr in cat_children.items()
                    }
                )

        family_tree: list[Any] = []
        for uid, child_type in children_list.items():
            recovered_object = self.get_entity(uid)[0]
            if recovered_object is None and not isinstance(entity, PropertyGroup):
                recovered_object = self.load_entity(uid, child_type, parent=entity)

            if recovered_object is None or isinstance(recovered_object, PropertyGroup):
                continue

            recovered_object.on_file = True
            recovered_object.entity_type.on_file = True
            family_tree += [recovered_object]

            if recursively and isinstance(recovered_object, (Group, ObjectBase)):
                family_tree += self.fetch_children(recovered_object, recursively=True)
                if (
                    isinstance(recovered_object, ObjectBase)
                    and recovered_object.property_groups is not None
                ):
                    family_tree += recovered_object.property_groups

        if isinstance(entity, ObjectBase) and entity.property_groups is not None:
            family_tree += entity.property_groups

        return family_tree

    def fetch_concatenated_attributes(
        self, entity: Concatenator | ConcatenatedObject
    ) -> dict | None:
        """
        Fetch attributes of Concatenated entity.
        :param entity: Concatenator group or ConcatenateObject.
        :return: Dictionary of attributes.
        """
        if isinstance(entity, Group):
            entity_type = "Group"
        else:
            raise NotImplementedError(
                "Method 'fetch_concatenated_attributes' currently only implemented "
                "for 'Group' entities."
            )

        return self._io_call(
            H5Reader.fetch_concatenated_attributes,
            entity.uid,
            entity_type,
            entity.concat_attr_str,
            mode="r",
        )

    def fetch_concatenated_list(
        self, entity: Group | ObjectBase, label: str
    ) -> list | None:
        """
        Fetch list of data or indices of ConcatenatedData entities.
        :param entity: Concatenator group.
        :param label: Label name of the h5py.Group.
        :return: List of concatenated Data names.
        """
        if isinstance(entity, Group):
            entity_type = "Group"
        else:
            raise NotImplementedError(
                "Method 'fetch_concatenated_list' currently only implemented "
                "for 'Group' entities."
            )

        return self._io_call(
            H5Reader.fetch_concatenated_attributes,
            entity.uid,
            entity_type,
            label,
            mode="r",
        )

    def fetch_concatenated_values(
        self, entity: Group | ObjectBase, label: str
    ) -> tuple | None:
        """
        Fetch data under the ConcatenatedData Data group of an entity.

        :param entity: Concatenator group.
        :param label: Name of the target data.

        :return: Index array and data values for the target label.
        """
        if isinstance(entity, Group):
            entity_type = "Group"
        else:
            raise NotImplementedError(
                "Method 'fetch_concatenated_values' currently only implemented "
                "for 'Group' entities."
            )

        return self._io_call(
            H5Reader.fetch_concatenated_values,
            entity.uid,
            entity_type,
            label,
            mode="r",
        )

    def fetch_metadata(self, entity: Entity, argument="Metadata") -> dict | None:
        """
        Fetch the metadata of an entity from the source geoh5.

        :param entity: Entity uid containing the metadata.
        :param argument: Optional argument for other json-like attributes.

        :return: Dictionary of values.
        """
        entity_type = str_from_type(entity)
        return self._io_call(
            H5Reader.fetch_metadata,
            entity.uid,
            argument=argument,
            entity_type=entity_type,
            mode="r",
        )

    def fetch_property_groups(self, entity: Entity) -> list[PropertyGroup]:
        """
        Fetch all property_groups on an object from the source geoh5

        :param entity: Target object

        :return: List of PropertyGroups
        """
        raise DeprecationWarning(
            f"Method 'fetch_property_groups' of {self} as been removed. "
            "Use `entity.property_groups` instead."
        )

    def fetch_type(self, uid: uuid.UUID, entity_type: str) -> dict:
        """
        Fetch attributes of a specific entity type.
        :param uid: Unique identifier of the entity type.
        :param entity_type: One of 'Data', 'Object' or 'Group'
        """
        return self._io_call(H5Reader.fetch_type, uid, entity_type)

    def fetch_values(self, entity: Entity) -> np.ndarray | str | float | None:
        """
        Fetch the data values from the source geoh5.

        :param entity: Entity with 'values'.

        :return: Array of values.
        """
        if isinstance(entity, (ConcatenatedObject | ConcatenatedData)):
            return entity.concatenator.fetch_values(entity, entity.name)

        return self._io_call(H5Reader.fetch_values, entity.uid)

    def fetch_file_object(self, uid: uuid.UUID, file_name: str) -> bytes | None:
        """
        Fetch an image from file name.
        :param uid: Unique identifier of target data object.
        :param file_name: Name of the file to fetch.
        :return: Array of values.
        """
        return self._io_call(H5Reader.fetch_file_object, uid, file_name)

    def finalize(self) -> None:
        """
        Deprecate method finalize.
        """
        warnings.warn(
            "The 'finalize' method will be deprecated in future versions of geoh5py in"
            " favor of `workspace.close()`. "
            "Please update your code to suppress this warning.",
            DeprecationWarning,
        )
        self.close()

    def find_data(self, data_uid: uuid.UUID) -> Entity | None:
        """
        Find an existing and active Data entity.
        """
        return weakref_utils.get_clean_ref(self._data, data_uid)

    def find_entity(self, entity_uid: uuid.UUID) -> Entity | PropertyGroup | None:
        """Get all active entities registered in the workspace."""
        return (
            self.find_group(entity_uid)
            or self.find_data(entity_uid)
            or self.find_object(entity_uid)
            or self.find_property_group(entity_uid)
        )

    def find_group(self, group_uid: uuid.UUID) -> Group | None:
        """
        Find an existing and active Group object.
        """
        return weakref_utils.get_clean_ref(self._groups, group_uid)

    def find_property_group(
        self, property_group_uid: uuid.UUID
    ) -> PropertyGroup | None:
        """
        Find an existing and active PropertyGroup object.
        """
        return weakref_utils.get_clean_ref(self._property_groups, property_group_uid)

    def find_object(self, object_uid: uuid.UUID) -> ObjectBase | None:
        """
        Find an existing and active Object.
        """
        return weakref_utils.get_clean_ref(self._objects, object_uid)

    def find_type(
        self, type_uid: uuid.UUID, type_class: type[EntityType]
    ) -> EntityType | None:
        """
        Find an existing and active EntityType.
        :param type_uid: Unique identifier of target type.
        :param type_class: The type of entity to find.
        """
        found_type = weakref_utils.get_clean_ref(self._types, type_uid)
        return found_type if isinstance(found_type, type_class) else None

    @property
    def ga_version(self) -> str:
        """
        :obj:`str` Version of Geoscience Analyst software.
        """
        return self._ga_version

    @ga_version.setter
    def ga_version(self, value: str):
        self._ga_version = value

    def get_entity(self, name: str | uuid.UUID) -> list[Entity | PropertyGroup | None]:
        """
        Retrieve an entity from one of its identifier, either by name or :obj:`uuid.UUID`.
        :param name: Object identifier, either name or uuid.
        :return: List of entities with the same given name.
        """
        if isinstance(name, uuid.UUID):
            list_entity_uid = [name]

        else:  # Extract all objects uuid with matching name
            list_entity_uid = [
                key for key, val in self.list_entities_name.items() if val == name
            ]

        if not list_entity_uid:
            return [None]

        entity_list: list[Entity | None | PropertyGroup] = []
        for uid in list_entity_uid:
            entity_list.append(self.find_entity(uid))

        return entity_list

    @property
    def groups(self) -> list[groups.Group]:
        """Get all active Group entities registered in the workspace."""
        return self._all_groups()

    @property
    def property_groups(self) -> list[PropertyGroup]:
        """Get all active PropertyGroup entities registered in the workspace."""
        return self._all_property_groups()

    @property
    def geoh5(self) -> h5py.File:
        """
        Instance of h5py.File.
        """
        if not self._geoh5:
            raise Geoh5FileClosedError

        return self._geoh5

    @property
    def h5file(self) -> str | Path | BytesIO | None:
        """
        Target *geoh5* file name with path or BytesIO object representation.

        On :func:`geoh5py.workspace.Workspace.save`, the BytesIO representation
        gets replaced by a Path to a file on disk.
        """
        return self._h5file

    @h5file.setter
    def h5file(self, file: str | Path | BytesIO | None):
        if self._h5file is not None:
            raise ValueError(
                "The 'h5file' attribute cannot be changed once it has been set."
            )

        if not isinstance(file, (str, Path, BytesIO, type(None))):
            raise ValueError(
                "The 'h5file' attribute must be a str, "
                "pathlib.Path to the target geoh5 file or BytesIO. "
                f"Provided {file} of type({type(file)})"
            )

        if isinstance(file, type(None)) or (
            isinstance(file, (str, Path)) and not Path(file).is_file()
        ):
            self._h5file = BytesIO()
            self._geoh5 = h5py.File(self.h5file, "a")

            with self._geoh5:
                self._root = self.create_root()
                H5Writer.init_geoh5(self.geoh5, self)

        elif isinstance(file, BytesIO):
            self._h5file = file

        if isinstance(file, (str, Path)):
            if Path(file).suffix != ".geoh5":
                raise ValueError("Input 'h5file' file must have a 'geoh5' extension.")

            if not Path(file).is_file():
                warnings.warn(
                    "From version 0.8.0, the 'h5file' attribute must be a string "
                    "or path to an existing file, or user must call the 'create' "
                    "method. We will attempt to `save` the file for you, but this "
                    "behaviour will be removed in future releases.",
                )
                self.save_as(file)
                self.close()
            else:
                self._h5file = Path(file)

    @property
    def list_data_name(self) -> dict[uuid.UUID, str]:
        """
        :obj:`dict` of :obj:`uuid.UUID` keys and name values for all registered Data.
        """
        data_name = {}
        for key, val in self._data.items():
            entity = val()
            if entity is not None:
                data_name[key] = entity.name
        return data_name

    @property
    def list_entities_name(self) -> dict[uuid.UUID, str]:
        """
        :return: :obj:`dict` of :obj:`uuid.UUID` keys and name values for all registered Entities.
        """
        entities_name = self.list_groups_name
        entities_name.update(self.list_objects_name)
        entities_name.update(self.list_data_name)
        entities_name.update(self.list_property_groups_name)

        return entities_name

    @property
    def list_groups_name(self) -> dict[uuid.UUID, str]:
        """
        :obj:`dict` of :obj:`uuid.UUID` keys and name values for all registered Groups.
        """
        groups_name = {}
        for key, val in self._groups.items():
            entity = val()
            if entity is not None:
                groups_name[key] = entity.name
        return groups_name

    @property
    def list_property_groups_name(self) -> dict[uuid.UUID, str]:
        """
        :obj:`dict` of :obj:`uuid.UUID` keys and name values for all registered Groups.
        """
        property_groups_name = {}
        for key, val in self._property_groups.items():
            entity = val()
            if entity is not None:
                property_groups_name[key] = entity.name
        return property_groups_name

    @property
    def list_objects_name(self) -> dict[uuid.UUID, str]:
        """
        :obj:`dict` of :obj:`uuid.UUID` keys and name values for all registered Objects.
        """
        objects_name = {}
        for key, val in self._objects.items():
            entity = val()
            if entity is not None:
                objects_name[key] = entity.name
                objects_name[key] = entity.name
        return objects_name

    def load_entity(
        self,
        uid: uuid.UUID,
        entity_type: str,
        parent: Entity | None = None,
    ) -> Entity | PropertyGroup | None:
        """
        Recover an entity from geoh5.
        :param uid: Unique identifier of entity
        :param entity_type: One of entity type 'group', 'object', 'data' or 'root'
        :param parent: Parent entity.
        :return entity: Entity loaded from geoh5
        """
        child = self.get_entity(uid)[0]
        if isinstance(child, (Entity, PropertyGroup)):
            return child

        base_classes = {
            "group": Group,
            "object": ObjectBase,
            "data": Data,
            "root": RootGroup,
        }

        entity_attrs, type_attrs, prop_groups = self._io_call(
            H5Reader.fetch_attributes, uid, entity_type, mode="r"
        )

        if entity_attrs is None:
            return None

        if parent is not None:
            entity_attrs["parent"] = parent
        try:
            entity = self.create_entity(
                base_classes[entity_type],
                save_on_creation=False,
                entity=entity_attrs,
                entity_type=type_attrs,
            )
        except TypeError as error:
            warnings.warn(
                f"Could not create an entity from the given attributes {type_attrs}. Skipping over."
                f"Error: {error}"
            )
            return None

        if isinstance(entity, ObjectBase) and len(prop_groups) > 0:
            for kwargs in prop_groups.values():
                entity.create_property_group(on_file=True, **kwargs)

        if entity is not None:
            entity.on_file = True

        return entity

    @property
    def name(self) -> str:
        """
        :obj:`str` Name of the project.
        """
        return self._name

    @property
    def objects(self) -> list[objects.ObjectBase]:
        """Get all active Object entities registered in the workspace."""
        return self._all_objects()

    def open(self, mode: str | None = None) -> Workspace:
        """
        Open a geoh5 file and load the tree structure.

        :param mode: Optional mode of h5py.File. Defaults to 'r+'.
        :return: `self`
        """
        if isinstance(self._geoh5, h5py.File) and self._geoh5:
            warnings.warn(f"Workspace already opened in mode {self._geoh5.mode}.")
            return self

        if mode is None:
            mode = self._mode

        try:
            self._geoh5 = h5py.File(self.h5file, mode)
        except OSError:
            self._geoh5 = h5py.File(self.h5file, "r")

        self._data = {}
        self._objects = {}
        self._groups = {}
        self._types = {}
        self._property_groups = {}

        proj_attributes = self._io_call(H5Reader.fetch_project_attributes, mode="r")

        for key, attr in proj_attributes.items():
            setattr(self, self._attribute_map[key], attr)

        self.fetch_or_create_root()

        return self

    def register(self, entity: Entity | EntityType | PropertyGroup):
        """
        Register an entity to the workspace based on its type.

        :param entity: The entity to be registered.
        """
        if isinstance(entity, EntityType):
            weakref_utils.insert_once(self._types, entity.uid, entity)
        elif isinstance(entity, Group):
            weakref_utils.insert_once(self._groups, entity.uid, entity)
        elif isinstance(entity, Data):
            weakref_utils.insert_once(self._data, entity.uid, entity)
        elif isinstance(entity, ObjectBase):
            weakref_utils.insert_once(self._objects, entity.uid, entity)
        elif isinstance(entity, PropertyGroup):
            weakref_utils.insert_once(self._property_groups, entity.uid, entity)
            if not entity.on_file:
                self.add_or_update_property_group(entity)
        else:
            raise ValueError(f"Entity of type {type(entity)} is not supported.")

    @property
    def root(self) -> RootGroup:
        """
        :obj:`~geoh5py.groups.root_group.RootGroup` entity.
        """
        return self._root

    @property
    def repack(self) -> bool:
        """
        Flag to repack the file after data deletion.
        """
        return self._repack

    @repack.setter
    def repack(self, value: bool):
        """
        :type value: bool.
        """
        self._repack = value

    def save(self, filepath: str | Path) -> Workspace:
        warnings.warn(
            "Workspace.save is deprecated and will be remove in future versions,"
            "use Workspace.save_as instead.",
            DeprecationWarning,
        )

        return self.save_as(filepath)

    def save_as(self, filepath: str | Path) -> Workspace:
        """
        Save the workspace to disk.
        """
        if self._geoh5 is not None:
            self.close()

        filepath = Path(filepath)

        if filepath.suffix == "":
            filepath = filepath.with_suffix(".geoh5")

        if filepath.suffix != ".geoh5":
            raise ValueError("Input 'h5file' file must have a 'geoh5' extension.")

        if filepath.exists():
            raise FileExistsError(f"File {filepath} already exists.")

        if isinstance(self.h5file, BytesIO):
            with open(filepath, "wb") as file:
                file.write(self.h5file.getbuffer())
        elif self.h5file is None:
            raise ValueError("Input 'h5file' file must be specified.")
        else:
            move(self.h5file, filepath, copy)

        self._h5file = filepath

        self.open()

        return self

    def save_entity(
        self,
        entity: Entity,
        add_children: bool = True,
        compression: int = 5,
    ) -> None:
        """
        Save or update an entity to geoh5.
        :param entity: Entity to be written to geoh5.
        :param add_children: Add children entities to geoh5.
        :param compression: Compression level for data.
        """
        if isinstance(entity, Concatenated):
            active_parent = self.get_entity(entity.concatenator.uid)[0]
            if not isinstance(active_parent, Concatenator):
                raise ValueError(
                    f"DrillholeGroup {entity.concatenator.name} is not registered in the "
                    "workspace. Please add it first."
                )
            active_parent.add_save_concatenated(entity)

            if hasattr(entity, "entity_type"):
                self.save_entity_type(entity.entity_type)

        else:
            self._io_call(
                H5Writer.save_entity,
                entity,
                add_children=add_children,
                mode="r+",
                compression=compression,
            )

    def save_entity_type(self, entity_type: EntityType) -> None:
        """
        Save or update an entity_type to geoh5.

        :param entity_type: Entity to be written to geoh5.
        """
        self._io_call(H5Writer.write_entity_type, entity_type, mode="r+")

    @property
    def types(self) -> list[EntityType]:
        """Get all active entity types registered in the workspace."""
        return self._all_types()

    def update_attribute(
        self,
        entity: Entity | EntityType | DataType,
        attribute: str,
        channel: str | None = None,
        **kwargs,
    ) -> None:
        """
        Save or update an entity to geoh5.

        :param entity: Entity to be written to geoh5.
        :param attribute: Name of the attribute to get updated to geoh5.
        :param channel: Optional channel argument for concatenated data and index.
        """
        if entity.on_file:
            if isinstance(entity, (ConcatenatedObject | ConcatenatedData)):
                entity.concatenator.update_attributes(entity, attribute)
            elif channel is not None:
                self._io_call(
                    H5Writer.update_concatenated_field,
                    entity,
                    attribute,
                    channel,
                    mode="r+",
                )
            else:
                self._io_call(
                    H5Writer.update_field, entity, attribute, mode="r+", **kwargs
                )

            self._io_call(H5Writer.clear_stats_cache, entity, mode="r+")

    def validate_data_type(self, attributes: dict, values) -> DataType:
        """
        Find or create a data type from input dictionary.

        :param attributes: Dictionary of attributes.
        :param values: Values to be stored as data.
        """
        entity_type = attributes.pop("entity_type", {})
        if isinstance(entity_type, DataType):
            if (entity_type.uid not in self._types) or (entity_type.workspace != self):
                entity_type = entity_type.copy(workspace=self)
        else:
            primitive_type = attributes.pop(
                "type", attributes.pop("primitive_type", None)
            )

            if primitive_type is None:
                primitive_type = DataType.primitive_type_from_values(values)

            if isinstance(primitive_type, str):
                primitive_type = DataType.validate_primitive_type(
                    primitive_type.upper()
                )

            # Generate a value map based on type of values
            if (
                primitive_type is PrimitiveTypeEnum.REFERENCED
                and "value_map" not in attributes
            ):
                attributes["value_map"] = values

            entity_type = DataType.find_or_create_type(
                self, primitive_type, **attributes
            )

        return entity_type

    @property
    def version(self) -> float:
        """
        :obj:`float` Version of the geoh5 file format.
        """
        return self._version

    @version.setter
    def version(self, value: float):
        self._version = value

    @property
    def workspace(self) -> Workspace:
        """
        This workspace instance itself.
        """
        return self

    def _io_call(self, fun, *args, mode="r", **kwargs):
        """
        Run a H5Writer or H5Reader function with validation of target geoh5
        """
        try:
            if self._geoh5 is None:
                return None

            if mode in ["r+", "a"] and self.geoh5.mode == "r":
                raise UserWarning(
                    f"Error performing {fun}. "
                    "Attempting to write to a geoh5 file in read-only mode. "
                    "Consider closing the workspace (Geoscience ANALYST) and "
                    "re-opening in mode='r+'."
                )

            return fun(self.geoh5, *args, **kwargs)

        except Geoh5FileClosedError as error:
            if not Path(str(self.h5file)).is_file() and not isinstance(
                self.h5file, BytesIO
            ):
                raise FileNotFoundError(
                    f"Error performing {fun}. "
                    "The geoh5 file does not exist."
                    r"Consider creating the geoh5 with Workspace().save('PATH\*.geoh5)'"
                ) from error

            raise Geoh5FileClosedError(
                f"Error executing {fun}. "
                + "Consider re-opening with `Workspace.open()' "
                "or used within a context manager."
            ) from error

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


@contextmanager
def active_workspace(workspace: Workspace):
    previous_active_ref = Workspace._active_ref  # pylint: disable=protected-access
    workspace.activate()
    yield workspace

    workspace.deactivate()
    # restore previous active workspace when leaving the context
    previous_active = previous_active_ref()
    if previous_active is not None:
        previous_active.activate()  # pylint: disable=no-member



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


# pylint: disable=too-few-public-methods

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any
from uuid import UUID

import numpy as np

from geoh5py import TYPE_UID_TO_CLASS, Workspace
from geoh5py.groups import Group, PropertyGroup
from geoh5py.objects import ObjectBase
from geoh5py.shared import Entity
from geoh5py.shared.exceptions import (
    AssociationValidationError,
    AtLeastOneValidationError,
    OptionalValidationError,
    PropertyGroupValidationError,
    RequiredValidationError,
    ShapeValidationError,
    TypeValidationError,
    UUIDValidationError,
    ValueValidationError,
    iterable,
)


def to_path(value: list[str]) -> list[Path]:
    """Promote path strings to patlib.Path objects."""
    out = []
    for path in value:
        if isinstance(path, str):
            out.append(Path(path))
        else:
            out.append(path)
    return out


def to_list(value: Any) -> list[Any]:
    """Promote single values to list."""
    if isinstance(value, str) and ";" in value:
        value = value.split(";")
    if not isinstance(value, list):
        value = [value]
    return value


def to_uuid(values):
    """Promote strings to uuid and pass anything else."""
    out = []
    for val in values:
        if isinstance(val, str):
            val = UUID(val)
        out.append(val)
    return out


def class_or_raise(value: UUID) -> type[ObjectBase] | type[Group]:
    """Promote uid to class, raise if uid is not a geoh5py type uid."""
    if value not in TYPE_UID_TO_CLASS:
        raise ValueError(
            f"Provided type_uid string {value!s} is not a recognized "
            f"geoh5py object or group type uid."
        )
    return TYPE_UID_TO_CLASS[value]


def to_class(
    values: list[UUID | type[ObjectBase] | type[Group]],
) -> list[type[ObjectBase] | type[Group]]:
    """
    Promote uid to class.

    Passes existing classes and raises if uid is not a geoh5py type uid.
    """
    out = []
    for val in values:
        if isinstance(val, UUID):
            out.append(class_or_raise(val))
        elif issubclass(val, (ObjectBase, Group)):
            out.append(val)
    return out


def empty_string_to_uid(value):
    """Promote empty string to uid, and pass all other values."""
    if value == "":
        return UUID("00000000-0000-0000-0000-000000000000")
    return value


class BaseValidator(ABC):
    """Concrete base class for validators."""

    validator_type: str

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)

    def __call__(self, *args):
        self.validate(*args)

    @classmethod
    @abstractmethod
    def validate(cls, name: str, value: Any, valid: Any):
        """
        Custom validation function.
        """
        raise NotImplementedError(
            "The 'validate' method must be implemented by the sub-class. "
            f"Must contain a 'name' {name}, 'value' {value} and 'valid' {valid} argument."
        )


class OptionalValidator(BaseValidator):
    """Validate that forms contain optional parameter if None value is given."""

    validator_type = "optional"

    @classmethod
    def validate(
        cls,
        name: str,
        value: Any | None,
        valid: bool,
    ) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: True if optional keyword in form for parameter.
        """
        if value is None and not valid:
            raise OptionalValidationError(name, value, valid)


class AssociationValidator(BaseValidator):
    """Validate the association between data and parent object."""

    validator_type = "association"

    @classmethod
    def validate(
        cls,
        name: str,
        value: Entity | PropertyGroup | UUID | None,
        valid: Entity | Workspace,
    ) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: Expected value shape
        """
        if valid is None:
            return

        if isinstance(valid, list):
            warnings.warn(
                "Data associated with multiSelect dependent is not supported. Validation ignored."
            )
            return

        if not isinstance(valid, (Entity, Workspace)):
            raise ValueError(
                "'AssociationValidator.validate' requires a 'valid'"
                " input of type 'Entity', 'Workspace' or None. "
                f"Provided '{valid}' of type {type(valid)} for parameter '{name}'"
            )

        if isinstance(value, UUID):
            uid = value
        elif isinstance(value, (Entity, PropertyGroup)):
            uid = value.uid
        else:
            return

        if isinstance(valid, Workspace):
            # TODO add a generic method to workspace to get all uuid
            children = valid.get_entity(uid)
            if None in children:
                children = valid.fetch_children(valid.root, recursively=True)

        elif isinstance(valid, Entity):
            children = valid.workspace.fetch_children(valid, recursively=True)

        if uid not in [getattr(child, "uid", None) for child in children]:
            raise AssociationValidationError(name, value, valid)


class PropertyGroupValidator(BaseValidator):
    """Validate property_group from parent entity."""

    validator_type = "property_group_type"

    @classmethod
    def validate(cls, name: str, value: PropertyGroup, valid: str | list[str]) -> None:
        if isinstance(valid, str):
            valid = [valid]

        if (value is not None) and (value.property_group_type not in valid):
            raise PropertyGroupValidationError(name, value, valid)


class AtLeastOneValidator(BaseValidator):
    validator_type = "one_of"

    @classmethod
    def validate(cls, name, value, valid):
        if not any(v for v in value.values()):
            raise AtLeastOneValidationError(name, value)


class RequiredValidator(BaseValidator):
    """
    Validate that required keys are present in parameter.
    """

    validator_type = "required"

    @classmethod
    def validate(cls, name: str, value: Any, valid: bool) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: Assert to be required
        """
        if value is None and valid:
            raise RequiredValidationError(name)


class ShapeValidator(BaseValidator):
    """Validate the shape of provided value."""

    validator_type = "shape"

    @classmethod
    def validate(cls, name: str, value: Any, valid: tuple[int, ...]) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: Expected value shape
        """
        if value is None:
            return

        if isinstance(value, np.ndarray):
            pshape = value.shape
        elif isinstance(value, list):
            pshape = (len(value),)
        else:
            pshape = (1,)

        if pshape != valid:
            raise ShapeValidationError(name, pshape, valid)


class TypeValidator(BaseValidator):
    """
    Validate the value type from a list of valid types.
    """

    validator_type = "types"

    @classmethod
    def validate(cls, name: str, value: Any, valid: type | list[type]) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: List of accepted value types
        """
        if isinstance(valid, type):
            valid = [valid]

        if not isinstance(valid, list):
            raise TypeError("Input `valid` options must be a type or list of types.")

        if not iterable(value) or (isinstance(value, list) and list in tuple(valid)):
            value = (value,)

        for val in value:
            if not isinstance(val, tuple(valid)):
                valid_names = [t.__name__ for t in valid if hasattr(t, "__name__")]
                type_name = type(val).__name__

                raise TypeValidationError(name, type_name, valid_names)


class UUIDValidator(BaseValidator):
    """Validate a uuui.UUID value or uuid string."""

    validator_type = "uuid"

    @classmethod
    def validate(cls, name: str, value: Any, valid: None = None) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter uuid.
        :param valid: [Optional] Validate uuid from parental entity or known uuids
        """

        if isinstance(value, str):
            try:
                value = UUID(value)
            except ValueError as exception:
                raise UUIDValidationError(name, str(value)) from exception


class ValueValidator(BaseValidator):
    """
    Validator that ensures that values are valid entries.
    """

    validator_type = "values"

    @classmethod
    def validate(cls, name: str, value: Any, valid: list[float | str]) -> None:
        """
        :param name: Parameter identifier.
        :param value: Input parameter value.
        :param valid: List of accepted values
        """
        if value is None:
            return

        if not isinstance(value, (list, tuple)):
            value = [value]

        for val in value:
            if val is not None and val not in valid:
                raise ValueValidationError(name, val, valid)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any
from uuid import UUID


if TYPE_CHECKING:
    from geoh5py import Workspace
    from geoh5py.groups import PropertyGroup
    from geoh5py.shared import Entity


class Geoh5FileClosedError(ABC, Exception):
    """Error for closed geoh5 file."""


class BaseValidationError(ABC, Exception):
    """Base class for custom exceptions."""

    @classmethod
    @abstractmethod
    def message(cls, name, value, validation):
        """Builds custom error message."""
        raise NotImplementedError()


class JSONParameterValidationError(Exception):
    """Error on uuid validation."""

    def __init__(self, name: str, err: str):
        super().__init__(JSONParameterValidationError.message(name, err))

    @classmethod
    def message(cls, name, err):
        return f"Malformed ui.json dictionary for parameter '{name}'. {err}"


class UIJsonFormatError(BaseValidationError):
    def __init__(self, name, msg):
        super().__init__(f"Invalid UIJson format for parameter '{name}'. {msg}")

    @classmethod
    def message(cls, name, value, validation):
        pass


class AggregateValidationError(BaseValidationError):
    def __init__(
        self,
        name: str,
        value: list[BaseValidationError],
    ):
        super().__init__(AggregateValidationError.message(name, value))

    @classmethod
    def message(cls, name, value, validation=None):
        msg = f"\n\nValidation of '{name}' collected {len(value)} errors:\n"
        for i, err in enumerate(value):
            msg += f"\t{i}. {err!s}\n"
        return msg


class OptionalValidationError(BaseValidationError):
    """Error if None value provided to non-optional parameter."""

    def __init__(
        self,
        name: str,
        value: Any | None,
        validation: bool,
    ):
        super().__init__(OptionalValidationError.message(name, value, validation))

    @classmethod
    def message(cls, name, value, validation):
        return f"Cannot set a None value to non-optional parameter: {name}."


class AssociationValidationError(BaseValidationError):
    """Error on association between child and parent entity validation."""

    def __init__(
        self,
        name: str,
        value: Entity | PropertyGroup | UUID,
        validation: Entity | Workspace,
    ):
        super().__init__(AssociationValidationError.message(name, value, validation))

    @classmethod
    def message(cls, name, value, validation):
        return (
            f"Property '{name}' with value: '{value}' must be "
            f"a child entity of parent {validation}"
        )


class PropertyGroupValidationError(BaseValidationError):
    """Error on property group validation."""

    def __init__(self, name: str, value: PropertyGroup, validation: list[str]):
        super().__init__(PropertyGroupValidationError.message(name, value, validation))

    @classmethod
    def message(cls, name, value, validation):
        return (
            f"Property group for '{name}' must be of type '{validation}'. "
            f"Provided '{value.name}' of type '{value.property_group_type}'"
        )


class AtLeastOneValidationError(BaseValidationError):
    def __init__(self, name: str, value: list[str]):
        super().__init__(AtLeastOneValidationError.message(name, value))

    @classmethod
    def message(cls, name, value, validation=None):
        opts = "'" + "', '".join(str(k) for k in value) + "'"
        return f"Must provide at least one {name}.  Options are: {opts}"


class TypeUIDValidationError(BaseValidationError):
    """Error on type uid validation."""

    def __init__(self, name: str, value, validation: list[str]):
        super().__init__(
            TypeUIDValidationError.message(
                name, value.default_type_uid(), list(validation)
            )
        )

    @classmethod
    def message(cls, name, value, validation):
        return (
            f"Type uid '{value}' provided for '{name}' is invalid."
            + iterable_message(validation)
        )


class RequiredValidationError(BaseValidationError):
    def __init__(self, name: str):
        super().__init__(RequiredValidationError.message(name))

    @classmethod
    def message(cls, name, value=None, validation=None):
        return f"Missing required parameter: '{name}'."


class InCollectionValidationError(BaseValidationError):
    collection = "Collection"
    item = "data"

    def __init__(self, name: str, value: list[str]):
        super().__init__(self.message(name, value))

    @classmethod
    def message(cls, name, value, validation=None):
        _ = validation
        return (
            f"{cls.collection}: '{name}' "
            f"is missing required {cls.item}(s): {value}."
        )


class RequiredFormMemberValidationError(InCollectionValidationError):
    collection = "Form"
    item = "member"


class RequiredUIJsonParameterValidationError(InCollectionValidationError):
    collection = "UIJson"
    item = "parameter"


class RequiredWorkspaceObjectValidationError(InCollectionValidationError):
    collection = "Workspace"
    item = "object"


class RequiredObjectDataValidationError(BaseValidationError):
    def __init__(self, name: str, value: list[tuple[str, str]]):
        super().__init__(self.message(name, value))

    @classmethod
    def message(cls, name, value, validation=None):
        _ = validation
        return (
            f"Workspace: '{name}' object(s) {[k[0] for k in value]} "
            f"are missing required children {[k[1] for k in value]}."
        )


class ShapeValidationError(BaseValidationError):
    """Error on shape validation."""

    def __init__(
        self, name: str, value: tuple[int, ...], validation: tuple[int, ...] | str
    ):
        super().__init__(ShapeValidationError.message(name, value, validation))

    @staticmethod
    def message(name, value, validation):
        return (
            f"Parameter '{name}' with shape {value} was provided. "
            f"Expected {validation}."
        )


class TypeValidationError(BaseValidationError):
    """Error on type validation."""

    def __init__(self, name: str, value: str, validation: str | list[str]):
        super().__init__(TypeValidationError.message(name, value, validation))

    @staticmethod
    def message(name, value, validation):
        return f"Type '{value}' provided for '{name}' is invalid." + iterable_message(
            validation
        )


class UUIDValidationError(BaseValidationError):
    """Error on uuid string validation."""

    def __init__(self, name: str, value: str):
        super().__init__(UUIDValidationError.message(name, value))

    @staticmethod
    def message(name, value, validation=None):
        return f"Parameter '{name}' with value '{value}' is not a valid uuid string."


class ValueValidationError(BaseValidationError):
    """Error on value validation."""

    def __init__(self, name: str, value: Any, validation: list[Any]):
        super().__init__(ValueValidationError.message(name, value, validation))

    @staticmethod
    def message(name, value, validation):
        return f"Value '{value}' provided for '{name}' is invalid." + iterable_message(
            validation
        )


def iterable_message(valid: list[Any] | None) -> str:
    """Append possibly iterable valid: "Must be (one of): {valid}."."""
    if valid is None:
        msg = ""
    elif iterable(valid, checklen=True):
        vstr = "'" + "', '".join(str(k) for k in valid) + "'"
        msg = f" Must be one of: {vstr}."
    else:
        msg = f" Must be: '{valid[0]}'."

    return msg


def iterable(value: Any, checklen: bool = False) -> bool:
    """
    Checks if object is iterable.

    Parameters
    ----------
    value : Object to check for iterableness.
    checklen : Restrict objects with __iter__ method to len > 1.

    Returns
    -------
    True if object has __iter__ attribute but is not string or dict type.
    """
    only_array_like = (not isinstance(value, str)) & (not isinstance(value, dict))
    if (hasattr(value, "__iter__")) & only_array_like:
        return not (checklen and (len(value) == 1))

    return False



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from typing import TYPE_CHECKING

from ..shared import EntityType


if TYPE_CHECKING:
    from ..workspace import Workspace
    from . import base


class GroupType(EntityType):
    _attribute_map = EntityType._attribute_map.copy()
    _attribute_map.update(
        {
            "Allow move contents": "allow_move_content",
            "Allow delete contents": "allow_delete_content",
        }
    )

    def __init__(
        self,
        workspace: Workspace,
        allow_move_content: bool = True,
        allow_delete_content: bool = True,
        **kwargs,
    ):
        super().__init__(workspace, **kwargs)

        self.allow_move_content = allow_move_content
        self.allow_delete_content = allow_delete_content

    @property
    def allow_move_content(self) -> bool:
        """
        Allow to move the group.
        """
        return self._allow_move_content

    @allow_move_content.setter
    def allow_move_content(self, allow: bool):
        if not isinstance(allow, bool) and allow != 1 and allow != 0:
            raise TypeError("'allow_move_content must be a boolean.")
        self._allow_move_content = bool(allow)

    @property
    def allow_delete_content(self) -> bool:
        """
        :obj:`bool`: [True] Allow to delete the group
        :obj:`~geoh5py.shared.entity.Entity.children`.
        """
        return self._allow_delete_content

    @allow_delete_content.setter
    def allow_delete_content(self, allow: bool):
        if not isinstance(allow, bool) and allow != 1 and allow != 0:
            raise TypeError("'allow_delete_content must be a boolean.")
        self._allow_delete_content = bool(allow)



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''

from __future__ import annotations

import uuid
import warnings
from io import BytesIO
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import TYPE_CHECKING, Any

import numpy as np
from PIL import Image
from PIL.TiffImagePlugin import TiffImageFile

from ..data import FilenameData
from ..shared.conversion import GeoImageConversion
from ..shared.utils import (
    PILLOW_ARGUMENTS,
    box_intersect,
    dip_points,
    xy_rotation_matrix,
)
from .object_base import ObjectBase


if TYPE_CHECKING:
    from ..objects import Grid2D


class GeoImage(ObjectBase):  # pylint: disable=too-many-public-methods
    """
    Image object class.

    The GeoImage object position is defined by the four corner vertices.
    The values displayed in the image are stored in a separate entity, called
    'GeoImageMesh_Image', and stored as 'GeoImage.image_data' attribute. The image
    values themselves can be accessed through the 'GeoImage.image' attribute.

    The 'image' data can be set with:
        - A File on disk
        - An array of values defining the pixels of the image
            - A 2D array of values will create a grayscale image.
            - A 3D array of values will create an RGB image
        - A PIL.Image object

    Setting the 'image' property will create a 'GeoImageMesh_Image' entity and remove
    the previous one.

    :param cells: Array of indices defining segments connecting vertices.
    :param dip: Dip of the image in degrees from the vertices position.
    :param image: Image data as a numpy array, PIL.Image, bytes, or path to an image file.
    :param rotation: Rotation of the image in degrees, counter-clockwise.
    :param vertices: Array of vertices defining the four corners of the image.
    """

    __VERTICES_DTYPE = np.dtype([("x", "<f8"), ("y", "<f8"), ("z", "<f8")])
    _TYPE_UID = uuid.UUID(
        fields=(0x77AC043C, 0xFE8D, 0x4D14, 0x81, 0x67, 0x75E300FB835A)
    )
    _converter: type[GeoImageConversion] = GeoImageConversion

    def __init__(
        self,
        *,
        cells: np.ndarray | list | tuple | None = None,
        dip: float | None = None,
        image: str | np.ndarray | BytesIO | Image.Image | FilenameData | None = None,
        rotation: float | None = None,
        vertices: np.ndarray | list | tuple | None = None,
        **kwargs,
    ):
        self._cells: np.ndarray | None
        self._vertices: np.ndarray | None
        self._tag: dict[int, Any] | None = None

        super().__init__(**kwargs)
        self._image_data: FilenameData | None = None
        self.vertices = vertices
        self.image = image
        self.cells = cells

        if rotation is not None:
            self.rotation = rotation

        if dip is not None:
            self.dip = dip

    @property
    def cells(self) -> np.ndarray:
        """
        Array of indices defining segments connecting vertices.
        """
        if self._cells is None and self.on_file:
            self._cells = self.workspace.fetch_array_attribute(self)

        return self._cells

    @cells.setter
    def cells(self, indices: np.ndarray | list | tuple | None):
        if isinstance(indices, (list, tuple)):
            indices = np.array(indices, ndmin=2)

        if indices is None:
            indices = np.c_[[0, 1, 2, 0], [0, 2, 3, 0]].T.astype("uint32")

        if not isinstance(indices, np.ndarray):
            raise TypeError(
                "Attribute 'cells' must be provided as type numpy.ndarray, list or tuple."
            )

        if indices.ndim != 2 or indices.shape != (2, 4):
            raise ValueError("Array of cells should be of shape (2, 4).")

        if not np.issubdtype(indices.dtype, np.integer):
            raise TypeError("Indices array must be of integer type")

        self._cells = indices

        if self.on_file:
            self.workspace.update_attribute(self, "cells")

    def copy(
        self,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        mask: np.ndarray | None = None,
        **kwargs,
    ):
        """
        Function to copy an entity to a different parent entity.

        :param parent: New parent for the copied object.
        :param copy_children: Copy children entities.
        :param clear_cache: Clear cache of data values.
        :param mask: Array of indices to sub-sample the input entity.
        :param kwargs: Additional keyword arguments.
        """
        if mask is not None:
            warnings.warn("Masking is not supported for GeoImage objects.")

        new_entity = super().copy(
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            **kwargs,
        )

        return new_entity

    def copy_from_extent(
        self,
        extent: np.ndarray,
        parent=None,
        *,
        copy_children: bool = True,
        clear_cache: bool = False,
        inverse: bool = False,
        **kwargs,
    ) -> GeoImage | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.copy_from_extent`.
        """
        # todo: save the temp grid in a temp workspace?
        if self.vertices is None:
            raise AttributeError("Vertices are not defined.")

        if self.image is None:
            warnings.warn("Image is not defined.")
            return None

        # transform the image to a grid
        grid = self.to_grid2d(parent=parent, mode="RGBA")

        # transform the image
        grid_transformed = grid.copy_from_extent(
            extent=extent,
            parent=parent,
            copy_children=copy_children,
            clear_cache=clear_cache,
            inverse=inverse,
            from_image=True,
            **kwargs,
        )

        if grid_transformed is None:
            grid.workspace.remove_entity(grid)
            warnings.warn("Image could not be cropped.")
            return None

        # transform the grid back to an image
        image_transformed = grid_transformed.to_geoimage(
            keys=grid_transformed.get_data_list(), mode="RGBA", normalize=False
        )

        grid.workspace.remove_entity(grid_transformed)
        grid.workspace.remove_entity(grid)

        return image_transformed

    @property
    def default_vertices(self) -> np.ndarray:
        """
        Assign the default vertices based on image pixel count
        """
        if self.image is not None:
            return np.asarray(
                [
                    [0, self.image.size[1], 0],
                    [self.image.size[0], self.image.size[1], 0],
                    [self.image.size[0], 0, 0],
                    [0, 0, 0],
                ]
            ).astype(float)
        return np.asarray(
            [
                [0, 1, 0],
                [1, 1, 0],
                [1, 0, 0],
                [0, 0, 0],
            ]
        ).astype(float)

    @property
    def dip(self) -> float:
        """
        Calculated dip of the image in degrees from the vertices position.
        :return: the dip angle.
        """
        # Get rotation matrix
        rotation_matrix = xy_rotation_matrix(np.deg2rad(-self.rotation))

        # Rotate the vertices
        rotated_vertices = (rotation_matrix @ self.vertices.T).T

        # Calculate the vector perpendicular to the rotation
        delta_xyz = rotated_vertices[0] - rotated_vertices[3]

        # Compute dip in degrees
        dip = np.rad2deg(
            np.arctan2(delta_xyz[2], np.sqrt(delta_xyz[0] ** 2 + delta_xyz[1] ** 2))
        )

        return dip

    @dip.setter
    def dip(self, new_dip: float):
        # Transform the vertices to a plane
        self.vertices = (
            dip_points(
                self.vertices - self.origin,
                np.deg2rad(new_dip - self.dip),
                np.deg2rad(self.rotation),
            )
            + self.origin
        )

    def georeference(self, reference: np.ndarray | list, locations: np.ndarray | list):
        """
        Georeference the image vertices (corners) based on input reference and
        corresponding world coordinates.

        :param reference: Array of integers representing the reference used as reference points.
        :param locations: Array of floats for the corresponding world coordinates
            for each input pixel.

        :return vertices: Corners (vertices) in world coordinates.
        """
        reference = np.asarray(reference)
        locations = np.asarray(locations)
        if self.image is None:
            raise AttributeError("An 'image' must be set before georeferencing.")

        if reference.ndim != 2 or reference.shape[0] < 3 or reference.shape[1] != 2:
            raise ValueError(
                "Input reference points must be a 2D array of shape(*, 2) "
                "with at least 3 control points."
            )

        if (
            locations.ndim != 2
            or reference.shape[0] != locations.shape[0]
            or locations.shape[1] != 3
        ):
            raise ValueError(
                "Input 'locations' must be a 2D array of shape(*, 3) "
                "with the same number of rows as the control points."
            )
        constant = np.ones(reference.shape[0])
        param_x, _, _, _ = np.linalg.lstsq(
            np.c_[constant, reference], locations[:, 0], rcond=None
        )
        param_y, _, _, _ = np.linalg.lstsq(
            np.c_[constant, reference], locations[:, 1], rcond=None
        )
        param_z, _, _, _ = np.linalg.lstsq(
            np.c_[constant, reference], locations[:, 2], rcond=None
        )

        corners = self.default_vertices[:, :2]

        self.vertices = np.c_[
            param_x[0] + corners @ param_x[1:],
            param_y[0] + corners @ param_y[1:],
            param_z[0] + corners @ param_z[1:],
        ]

        self.set_tag_from_vertices()

    def georeferencing_from_image(self):
        """
        Georeferencing the GeoImage from the image.
        """
        if self.image is not None:
            if self.tag is not None:
                self.vertices = self.default_vertices
                self.georeferencing_from_tiff()
            else:
                self.vertices = self.default_vertices

    def georeferencing_from_tiff(self):
        """
        Get the geographic information from the PIL Image to georeference it.
        Run the georeference() method of the object.
        """
        if self.tag is None:
            raise AttributeError("The image is not georeferenced")

        try:
            # get geographic information
            u_origin = float(self.tag[33922][3])
            v_origin = float(self.tag[33922][4])
            u_cell_size = float(self.tag[33550][0])
            v_cell_size = float(self.tag[33550][1])
            u_count = float(self.tag[256][0])
            v_count = float(self.tag[257][0])
            u_oposite = float(u_origin + u_cell_size * u_count)
            v_oposite = float(v_origin - v_cell_size * v_count)

            # prepare georeferencing
            reference = np.array([[0.0, v_count], [u_count, v_count], [u_count, 0.0]])

            locations = np.array(
                [
                    [u_origin, v_origin, 0.0],
                    [u_oposite, v_origin, 0.0],
                    [u_oposite, v_oposite, 0.0],
                ]
            )

            # georeference the raster
            self.georeference(reference, locations)
        except KeyError:
            warnings.warn("The 'tif.' image has no referencing information.")

    @property
    def image(self):
        """
        Get the image as a :obj:`PIL.Image` object.
        """
        if self.image_data is not None and self.image_data.file_bytes is not None:
            return Image.open(BytesIO(self.image_data.file_bytes))

        return None

    @image.setter
    def image(
        self, image: str | np.ndarray | BytesIO | Image.Image | FilenameData | None
    ):
        if self._image_data is not None:
            raise AttributeError(
                "The 'image' property cannot be reset. "
                "Consider creating a new object."
            )

        if isinstance(image, (FilenameData, type(None))):
            self._image_data = image
            return

        image = self.validate_image_data(image)

        with TemporaryDirectory() as tempdir:
            if image.mode not in PILLOW_ARGUMENTS:
                raise NotImplementedError(
                    f"The mode {image.mode} of the image is not supported."
                )

            temp_file = Path(tempdir) / "image"
            image.save(temp_file, **PILLOW_ARGUMENTS[image.mode])
            image_file = self.add_file(str(temp_file))
            image_file.name = "GeoImageMesh_Image"
            image_file.entity_type.name = "GeoImageMesh_Image"

        self._image_data = image_file

        if self._vertices is None:
            self.vertices = self.default_vertices

        # if the image is a tiff save tag information
        if isinstance(image, TiffImageFile):
            self.tag = image
            self.to_grid2d(name=self.name + "_grid2d")

    @property
    def image_data(self) -> FilenameData | None:
        """
        Get the FilenameData entity holding the image.
        """
        if self._image_data is None:
            for child in self.children:
                if (
                    isinstance(child, FilenameData)
                    and child.name == "GeoImageMesh_Image"
                ):
                    self._image_data = child

        return self._image_data

    @property
    def image_georeferenced(self) -> Image.Image | None:
        """
        Get the image as a georeferenced :obj:`PIL.Image` object.
        """
        if self.tag is not None and self.image is not None:
            image = self.image

            # modify the exif
            for id_ in self.tag:
                image.getexif()[id_] = self.tag[id_]

            return image

        return None

    def mask_by_extent(
        self, extent: np.ndarray, inverse: bool = False
    ) -> np.ndarray | None:
        """
        Sub-class extension of :func:`~geoh5py.shared.entity.Entity.mask_by_extent`.

        Uses the four corners of the image to determine overlap with the extent window.
        """
        if self.extent is None or not box_intersect(self.extent, extent):
            return None

        return np.ones(self.vertices.shape[0], dtype=bool)

    @property
    def n_cells(self):
        """
        Number of vertices
        """
        return self.cells.shape[0]

    @property
    def n_vertices(self):
        """
        Number of vertices
        """
        return self.vertices.shape[0]

    @property
    def origin(self) -> np.array:
        """
        The origin of the image.
        :return: an array of the origin of the image in x, y, z.
        """

        return self.vertices[3, :]

    @property
    def rotation(self) -> float:
        """
        The rotation of the image in degrees, counter-clockwise.
        :return: the rotation angle.
        """
        dxy = np.r_[np.diff(self.vertices[:2, 0]), np.diff(self.vertices[:2, 1])]
        dxy /= np.linalg.norm(dxy)
        rotation_rad = np.arctan2(dxy[1], dxy[0])

        return np.rad2deg(rotation_rad)

    @rotation.setter
    def rotation(self, new_rotation):
        # Compute rotation matrix
        rotation_matrix = xy_rotation_matrix(np.deg2rad(new_rotation - self.rotation))

        # get the vertices without the origin
        vertices = self.vertices - self.origin

        # get the rotation matrix
        vertices = rotation_matrix @ vertices.T

        # save the vertices
        self.vertices = vertices.T + self.origin

    def save_as(self, name: str, path: str | Path = ""):
        """
        Function to save the geoimage into an image file.
        It the name ends by '.tif' or '.tiff' and the tag is not None
        then the image is saved as georeferenced tiff image ;
        else, the image is save with PIL.Image's save function.
        :param name: the name to give to the image.
        :param path: the path of the file of the image, default: ''.
        """
        # verifications
        if self.image is None:
            raise AttributeError("The object contains no image data")
        if not isinstance(name, str):
            raise TypeError(
                f"The 'name' has to be a string; a '{type(name)}' was entered instead"
            )
        if not isinstance(path, (str, Path)):
            raise TypeError(
                f"The 'path' has to be a string or a Path; a '{type(name)}' was entered instead"
            )
        if path != "" and not Path(path).is_dir():
            raise FileNotFoundError(f"No such file or directory: {path}")

        if name.endswith((".tif", ".tiff")) and self.tag is not None:
            # save the image
            image: Image = self.image_georeferenced
            image.save(Path(path) / name, exif=image.getexif())
        else:
            self.image.save(Path(path) / name)

    def set_tag_from_vertices(self):
        """
        If tag is None, set the basic tag values based on vertices
        in order to export as a georeferenced .tiff.
        WARNING: this function must be used after georeference().
        """
        if self._tag is None:
            self._tag = {}

        if self.image is None:
            raise AttributeError("An 'image' must be set before georeferencing.")

        width, height = self.image.size
        self._tag[256] = (width,)
        self._tag[257] = (height,)
        self._tag[33922] = (
            0.0,
            0.0,
            0.0,
            self.vertices[0, 0],
            self.vertices[0, 1],
            self.vertices[0, 2],
        )
        self._tag[33550] = (
            abs(self.vertices[1, 0] - self.vertices[0, 0]) / width,
            abs(self.vertices[0, 1] - self.vertices[2, 1]) / height,
            0.0,
        )

    @property
    def tag(self) -> dict | None:
        """
        Georeferencing information of a tiff image stored in the header.
        :return: a dictionary containing the PIL.Image.tag information.
        """
        if self._tag:
            return self._tag.copy()
        return None

    @tag.setter
    def tag(self, value: Image.Image | dict | None):
        if isinstance(value, (Image.Image, TiffImageFile)):
            self._tag = dict(value.tag)
        elif isinstance(value, dict):
            self._tag = value
        elif value is None:
            self._tag = None
        else:
            raise ValueError("Input 'tag' must be a PIL.Image")

    def to_grid2d(
        self,
        mode: str | None = None,
        **grid2d_kwargs,
    ) -> Grid2D:
        """
        Create a geoh5py :obj:geoh5py.objects.grid2d.Grid2D from the geoimage in the same workspace.

        :param mode: The output image mode, defaults to the incoming image.mode.
            If "GRAY" convert the image to grayscale.
        :param grid2d_kwargs: Keyword arguments to pass to the
            :obj:`geoh5py.objects.grid2d.Grid2D` constructor.

        :return: the new created :obj:`geoh5py.objects.grid2d.Grid2D`.
        """
        return self.converter.to_grid2d(self, mode, **grid2d_kwargs)

    def validate_image_data(
        self, image: str | np.ndarray | BytesIO | Image.Image | FilenameData | None
    ) -> Image.Image:
        """
        Validate the input image.

        :param image: Image to validate.

        :return: Image converted to FileNameData object.
        """
        # todo: this should be changed in the future to accept tiff images
        if isinstance(image, np.ndarray) and image.ndim in [2, 3]:
            if image.ndim == 3 and image.shape[2] != 3:
                raise ValueError(
                    "Shape of the 'image' must be a 2D or "
                    "a 3D array with shape(*,*, 3) representing 'RGB' values."
                )
            value = image
            if image.min() < 0 or image.max() > 255 or image.dtype != "uint8":
                value = image.astype(float)
                value -= value.min()
                value *= 255.0 / value.max()
                value = value.astype("uint8")

            image = Image.fromarray(value)

        elif isinstance(image, str):
            if not Path(image).is_file():
                raise ValueError(f"Input image file {image} does not exist.")

            image = Image.open(image)
        elif isinstance(image, bytes):
            image = Image.open(BytesIO(image))

        elif not isinstance(image, Image.Image):
            raise ValueError(
                "Input 'value' for the 'image' property must be "
                "a 2D or 3D numpy.ndarray, bytes, PIL.Image or a path to an existing image."
                f"Get type {type(image)} instead."
            )

        return image

    @property
    def vertices(self) -> np.ndarray:
        """
        :obj:`~geoh5py.objects.object_base.ObjectBase.vertices`:
        Defines the four corners of the geo_image
        """
        # Case the vertices were removed from the object
        if self._vertices is None and self.on_file:
            self._vertices = self.workspace.fetch_array_attribute(self, "vertices")

        # Case where the vertices are not set but the image is defined
        if self._vertices is None and self.tag is not None and self.image is not None:
            self.vertices = self.default_vertices
            self.georeferencing_from_tiff()

        # Case neither vertices nor image are set
        if self._vertices is None:
            return self.default_vertices

        return self._vertices.view("<f8").reshape((-1, 3)).astype(float)

    @vertices.setter
    def vertices(self, xyz: np.ndarray | list | None):
        if xyz is None:
            self._vertices = None
            return

        if isinstance(xyz, list | tuple):
            xyz = np.array(xyz, ndmin=2)

        if not isinstance(xyz, np.ndarray):
            raise TypeError(
                "Input 'vertices' must be provided as type numpy.ndarray, list or tuple."
            )

        if np.issubdtype(xyz.dtype, np.number):
            xyz = np.asarray(
                np.core.records.fromarrays(xyz.T, dtype=self.__VERTICES_DTYPE)
            )

        if xyz.dtype != self.__VERTICES_DTYPE:
            raise TypeError(
                f"Array of 'vertices' must be of dtype = {self.__VERTICES_DTYPE}"
            )

        if len(xyz) != 4:
            raise ValueError("Array of 'vertices' must be of shape (4, 3).")

        self._vertices = xyz
        self._tag = None

        if self.on_file:
            self.workspace.update_attribute(self, "vertices")



# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
#  Copyright (c) 2025 Mira Geoscience Ltd.                                     '
#                                                                              '
#  This file is part of geoh5py.                                               '
#                                                                              '
#  geoh5py is free software: you can redistribute it and/or modify             '
#  it under the terms of the GNU Lesser General Public License as published by '
#  the Free Software Foundation, either version 3 of the License, or           '
#  (at your option) any later version.                                         '
#                                                                              '
#  geoh5py is distributed in the hope that it will be useful,                  '
#  but WITHOUT ANY WARRANTY; without even the implied warranty of              '
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               '
#  GNU Lesser General Public License for more details.                         '
#                                                                              '
#  You should have received a copy of the GNU Lesser General Public License    '
#  along with geoh5py.  If not, see <https://www.gnu.org/licenses/>.           '
# ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''


from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Annotated, Any
from uuid import UUID

import numpy as np
from pydantic import (
    BaseModel,
    ConfigDict,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.alias_generators import to_camel
from pydantic.functional_validators import BeforeValidator

from geoh5py.groups import Group
from geoh5py.objects import ObjectBase
from geoh5py.shared.validators import (
    empty_string_to_uid,
    to_class,
    to_list,
    to_path,
    to_uuid,
)
from geoh5py.ui_json.validation import UIJsonError


class DependencyType(str, Enum):
    ENABLED = "enabled"
    DISABLED = "disabled"


class BaseForm(BaseModel):
    """
    Base class for uijson forms

    :param label: Label for ui element.
    :param value: The parameter's value.
    :param optional: If True, ui element is rendered with a checkbox to
        control the enabled state.
    :param enabled: If False, ui element is rendered grey and value will
        be written to file as None.
    :param main: Controls whether ui element will render in the general
        parameters tab (True) or optional parameters (False).
    :param tooltip: String rendered on hover over ui element.
    :param group: Grouped ui elements will be rendered within a box labelled
        with the group name.
    :param group_optional: If True, ui group is rendered with a checkbox that
        controls the enabled state of all of the groups members
    :param dependency: Name of parameter that controls the enabled or
        visible state of the ui element.
    :param dependency_type: Controls whether the ui element is enabled
        or visible when the dependency is enabled if optional or True
        if a bool type.
    :param group_dependency: Name of parameter that controls the enabled
        or visible state of the ui group.
    :param group_dependency_type: Controls whether the ui group is
        enabled or visible when the group dependency is enabled if
        optional or True if a bool type.
    """

    model_config = ConfigDict(
        extra="allow",
        frozen=True,
        populate_by_name=True,
        loc_by_alias=True,
        alias_generator=to_camel,
    )

    label: str
    value: Any
    optional: bool = False
    enabled: bool = True
    main: bool = True
    tooltip: str = ""
    group: str = ""
    group_optional: bool = False
    dependency: str = ""
    dependency_type: DependencyType = DependencyType.ENABLED
    group_dependency: str = ""
    group_dependency_type: DependencyType = DependencyType.ENABLED

    @property
    def json_string(self):
        return self.model_dump_json(exclude_unset=True, by_alias=True)

    def flatten(self):
        """Returns the data for the form."""
        return self.value

    def validate_data(self, params: dict[str, Any]):
        """Validate the form data."""


class StringForm(BaseForm):
    """
    String valued uijson form.
    """

    value: str = ""


class BoolForm(BaseForm):
    """
    Boolean valued uijson form.
    """

    value: bool = True


class IntegerForm(BaseForm):
    """
    Integer valued uijson form.
    """

    value: int = 1
    min: float = -np.inf
    max: float = np.inf


class FloatForm(BaseForm):
    """
    Float valued uijson form.
    """

    value: float = 1.0
    min: float = -np.inf
    max: float = np.inf
    precision: int = 2
    line_edit: bool = True


class ChoiceForm(BaseForm):
    """
    Choice list uijson form.
    """

    value: list[str]
    choice_list: list[str]
    multi_select: bool = False

    @field_validator("value", mode="before")
    @classmethod
    def to_list(cls, value):
        if not isinstance(value, list):
            value = [value]
        return value

    @field_serializer("value", when_used="json")
    def string_if_single(self, value):
        if len(value) == 1:
            value = value[0]
        return value

    @model_validator(mode="after")
    def valid_choice(self):
        bad_choices = []
        for val in self.value:
            if val not in self.choice_list:
                bad_choices.append(val)
        if bad_choices:
            raise ValueError(f"Provided value(s): {bad_choices} are not valid choices.")

        return self


PathList = Annotated[
    list[Path],
    BeforeValidator(to_path),
    BeforeValidator(to_list),
]


class FileForm(BaseForm):
    """
    File path uijson form
    """

    value: PathList
    file_description: list[str]
    file_type: list[str]
    file_multi: bool = False

    @field_serializer("value", when_used="json")
    def to_string(self, value):
        return ";".join([str(path) for path in value])

    @field_validator("value")
    @classmethod
    def valid_file(cls, value):
        bad_paths = []
        for path in value:
            if not path.exists():
                bad_paths.append(path)
        if any(bad_paths):
            raise ValueError(f"Provided path(s) {bad_paths} does not exist.")
        return value

    @model_validator(mode="after")
    def same_length(self):
        if len(self.file_description) != len(self.file_type):
            raise ValueError("File description and type lists must be the same length.")
        return self

    @model_validator(mode="before")
    @classmethod
    def value_file_type(cls, data):
        bad_paths = []
        for path in data["value"].split(";"):
            if Path(path).suffix[1:] not in data["file_type"]:
                bad_paths.append(path)
        if any(bad_paths):
            raise ValueError(f"Provided paths {bad_paths} have invalid extensions.")
        return data


MeshTypes = Annotated[
    list[type[ObjectBase] | type[Group]],
    BeforeValidator(to_class),
    BeforeValidator(to_uuid),
    BeforeValidator(to_list),
]


class ObjectForm(BaseForm):
    """
    Geoh5py object uijson form.
    """

    value: UUID = UUID("00000000-0000-0000-0000-000000000000")
    mesh_type: MeshTypes

    _empty_string_to_uid = field_validator("value", mode="before")(empty_string_to_uid)


class Association(str, Enum):
    """
    Geoh5py object association types.
    """

    VERTEX = "Vertex"
    CELL = "Cell"
    FACE = "Face"


class DataType(str, Enum):
    """
    Geoh5py data types.
    """

    INTEGER = "Integer"
    FLOAT = "Float"
    BOOLEAN = "Boolean"
    REFERENCED = "Referenced"
    VECTOR = "Vector"
    DATETIME = "DateTime"
    GEOMETRIC = "Geometric"
    TEXT = "Text"


class DataForm(BaseForm):
    """
    Geoh5py data uijson form.
    """

    value: UUID | float | int
    parent: str
    association: Association | list[Association]
    data_type: DataType | list[DataType]
    is_value: bool = False
    property: UUID = UUID("00000000-0000-0000-0000-000000000000")
    min: float = -np.inf
    max: float = np.inf
    precision: int = 2

    @field_validator("property", mode="before")
    @classmethod
    def empty_string_to_uid(cls, val):
        if val == "":
            return UUID("00000000-0000-0000-0000-000000000000")
        return val

    @model_validator(mode="after")
    def value_if_is_value(self):
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and self.is_value
        ):
            if isinstance(self.value, UUID):
                raise ValueError("Value must be numeric if is_value is True.")
        return self

    @model_validator(mode="after")
    def property_if_not_is_value(self):
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and "property" not in self.model_fields_set  # pylint: disable=unsupported-membership-test
        ):
            raise ValueError("A property must be provided if is_value is used.")
        return self

    def _validate_parent(self, params: dict[str, Any]):
        """Validate form uid is a child of the parent object."""
        child = None
        if isinstance(self.value, UUID):
            child = self.value
        elif "property" in list(self.model_fields_set) and not self.is_value:
            child = self.property

        if child is not None:
            if (
                not isinstance(params[self.parent], ObjectBase)
                or params[self.parent].get_entity(child)[0] is None
            ):
                raise UIJsonError(f"{child} data is not a child of {self.parent}.")

    def validate_data(self, params: dict[str, Any]):
        """Validate the form data."""
        self._validate_parent(params)

    def flatten(self):
        """Returns the data for the form."""
        if (
            "is_value" in self.model_fields_set  # pylint: disable=unsupported-membership-test
            and not self.is_value
        ):
            return self.property
        return self.value
